#include <string.h>
#include "geosim.h"
#include "gspriv.h"
#include "ietypes.h"
#include "windfunc.h"
#include "objprop.h"
#include "windprop.h"
#include "editrec.h"
#include "ifw.h"
#include "iemisc.h"

#ifndef MSDOS
#include <stdlib.h>
#endif

/*extern char *IEColorlist[], *IEAlias[];*/
extern char *Colorlist[], *Alias[];
extern boolean objwind;
extern GSColor redCol, blackCol;
extern boolean twindVisible;
extern IEIconfig iestate;

/* Internal protos */
void DelCurrentObject();
void ShowPropWind();
void SetActiveButtonText();
void DisplayActiveButton();
void SetLabPosButtonText();
void DisplayLabPosButton();
static void ObjSetupColorList();
static void ObjSetupFuncList();
static void EnableColorButtons();
static void EnableColor1();
static void DisableColor1();
static void DisableColorButtons();
void ShowColorButtonFields(int which);
static char *GetColor(int);
/**/

#define F_ALL 4
#define F_COLOR1 0
#define F_COLOR2 1
#define F_COLOR3 2
#define F_COLOR4 3

typedef struct {
	  unsigned int windFlags;
	  int numColors;
	  char *windName;
	  char *Bool1False, *Bool1True;
	  char *Bool2False, *Bool2True;
	  } WindParams;
	  
#define NUM_PROP_WINDS 7
#define FIELD_W 0
#define DRAGAREA_W 1
#define RECT_W 2
#define LINE_W 3
#define LABEL_W 4
#define BUTTON_W 5
#define POPUP_W 6
 
#define COLORLIST 1
#define CLIST_HAS_NONE 1 << 1
#define FUNCLIST 1 << 2
#define FUNCREC 1 << 3
#define WD_HT 1 << 4
#define HASTEXT 1 << 5
#define NAME 1 << 6
#define STATUS_TYPE 1 << 7
#define BOOL1 1 << 8
#define BOOL2 1 << 9
#define ACTIVE 1 << 10
#define LABPOS 1 << 11

#define FIELD_F (COLORLIST) | (WD_HT) | (NAME)/* | (COLOR1)*/
#define DRAGAREA_F (COLORLIST) | (WD_HT) | (NAME) | (BOOL1) | (FUNCREC) /*| (COLOR1)*/
#define RECT_F (COLORLIST) | (WD_HT) | (BOOL1) | (BOOL2) /*| (COLOR1)*/
#define LINE_F (COLORLIST) | (WD_HT) | (BOOL1) /*| (COLOR1)*/
#define LABEL_F (COLORLIST) | (CLIST_HAS_NONE) | (HASTEXT) | (BOOL1)/* | (COLOR1) | (COLOR2)*/
#define BUTTON_F (COLORLIST) | (FUNCLIST) | (WD_HT) | (NAME) | \
		(HASTEXT) | (STATUS_TYPE) | (BOOL1) | (BOOL2) | (ACTIVE) /*| (COLOR1) | (COLOR2) | (COLOR3)*/ | \
		(LABPOS) | (FUNCREC)
#define POPUP_F (COLORLIST) | (NAME) /*| (COLOR1) | (COLOR2) | (COLOR3) | (COLOR4)*/ | (BOOL1)

WindParams objParams[NUM_PROP_WINDS] = {
	{FIELD_F,    1, "fieldprop", NULL, NULL, NULL, NULL},
	{DRAGAREA_F, 1, "dragareaprop", "Inactive", "Active", NULL, NULL},
	{RECT_F,     1, "rectprop", "Thin", "Thick", "No Fill", "Fill"},
	{LINE_F,     1, "lineprop", "Thin", "Thick", NULL, NULL},
	{LABEL_F,    2, "labelprop", "Small", "Large", NULL, NULL},
	{BUTTON_F,   3, "buttonprop", "No Border", "Border", "Not Repeatable", "Repeatable"},
	{POPUP_F,	 4, "popupprop", "No Checklist", "Checklist", NULL, NULL}
	};
	
static PropertyRec objProp;
static int curWindNum = NUM_PROP_WINDS;
static void *curObj;
static EditRec *name, *text, *func;
static GSsclistHandle colorScroll, funcScroll;
static GSlistHandle colorList, funcList;
static int numListColors, numColors, curColor, numFuncs, curFunc;
static boolean cancelDel;
static int objbounds_x, objbounds_y, objbounds_wd, objbounds_ht; /* window relative */

static void objTextFunc(GSlistHandle list, int slot)
{	
	if ( ((list == colorList)&&(slot == curColor))	||
		((list == funcList)&&(slot == curFunc)) )
		GSsetcolor(GSconvertcolor("red"));
	else
		GSsetcolor(GSconvertcolor("black"));
	GSsetfont(SMALL);
}

static char *objLabelFunc(int slot)
{
/*	if (slot < numListColors)
		return Colorlist[slot];
	else if (slot < numColors)
		return Alias[slot-numListColors];
	else	
		return "";*/
	return GetColor(slot);
}

static void objInFunc() 
{
}

static void objOutFunc() 
{
}

static void objClickFunc(int slot)
{
	int old = curColor;
	
	curColor = slot;
	
	if (old == NO_ITEM)
		EnableColorButtons();
	
	/* also, make sure color1 isn't enabled for "NONE" */
	
	if (((old == NO_ITEM)||(old == numColors)) && (slot != numColors))
		EnableColor1();
	else if (!((old == NO_ITEM)||(old == numColors)) && (slot == numColors))
		DisableColor1();
	
	GSdrawlistitem(colorList, New, old);
	GSdrawlistitem(colorList, New, curColor);
}

static char *fnLabelFunc(int slot)
{
	IEPopup *p = GetPopupNum(slot);
	
	if (p != NULL)
		return p->title;
	else
		return "Error!";
}

static void fnInFunc()
{
}

static void fnOutFunc()
{
}

static void fnClickFunc(int slot)
{
	IEPopup *p = GetPopupNum(slot);
	int old = curFunc;
	int wflags = objParams[curWindNum].windFlags;

	curFunc = slot;
	GSdrawlistitem(funcList, New, old);
	GSdrawlistitem(funcList, New, curFunc);
	
	if ((p != NULL) && ((wflags) & (FUNCREC))) {
		sprintf(objProp.funcName, "@%s", p->title);
		UpdateEditRec(func);
		}
}

void ButtonToProp(IEButton *bt, IEWindow *wind)
{
	curObj = bt;
	curWindNum = BUTTON_W;
	
	strcpy(objProp.name, bt->name);
	strcpy(objProp.text, bt->text);
	objbounds_x = objProp.x = bt->x;
	objbounds_y = objProp.y = bt->y;
	objbounds_wd = objProp.wd = bt->wd;
	objbounds_ht = objProp.ht = bt->ht;
	strcpy(objProp.color[0], bt->efgcolorname);
	strcpy(objProp.color[1], bt->dfgcolorname);
	strcpy(objProp.color[2], bt->bgcolor);
	strcpy(objProp.funcName, bt->funcname);
	objProp.active = bt->active;
	objProp.bool1 = bt->border;
	objProp.bool2 = bt->repeatable;
	objProp.labpos = bt->textjust;
	objProp.objWind = wind;
}

void PropToButton()
{
	IEButton *bt = (IEButton *) curObj;
	int t1, t2;
	
	strcpy(bt->name, objProp.name);
	strcpy(bt->text, objProp.text);

	objbounds_x = MIN(objProp.x, bt->x);
	objbounds_y = MIN(objProp.y, bt->y);
	
	t1 = bt->x+bt->wd;
	t2 = objProp.x+objProp.wd;
	objbounds_wd = MAX(t1, t2) - objbounds_x;
	t1 = bt->y+bt->ht;
	t2 = objProp.y+objProp.ht;
	objbounds_ht = MAX(t1, t2) - objbounds_y;

	bt->x = objProp.x;
	bt->y = objProp.y;
	bt->wd = objProp.wd;
	bt->ht = objProp.ht;
	strcpy(bt->efgcolorname, objProp.color[0]);
	strcpy(bt->dfgcolorname, objProp.color[1]);
	strcpy(bt->bgcolor, objProp.color[2]);
	strcpy(bt->funcname, objProp.funcName);
	bt->active = objProp.active;
	bt->border = objProp.bool1;
	bt->textjust = objProp.labpos;
	bt->repeatable = objProp.bool2;
}

void FieldToProp(IEField *fld, IEWindow *wind)
{
	curObj = fld;
	curWindNum = FIELD_W;
	
	strcpy(objProp.name, fld->name);
	objbounds_x = objProp.x = fld->x;
	objbounds_y = objProp.y = fld->y;
	objbounds_wd = objProp.wd = fld->wd;
	objbounds_ht = objProp.ht = fld->ht;
	strcpy(objProp.color[0], fld->col);
	objProp.objWind = wind;
}

void PropToField()
{
	IEField *fld = (IEField *) curObj;
	int t1, t2;
		
	strcpy(fld->name, objProp.name);
	
	objbounds_x = MIN(fld->x, objProp.x);
	objbounds_y = MIN(fld->y, objProp.y);
	
	t1 = fld->x+fld->wd;
	t2 = objProp.x+objProp.wd;
	objbounds_wd = MAX(t1, t2) - objbounds_x;
	t1 = fld->y+fld->ht;
	t2 = objProp.y+objProp.ht;
	objbounds_ht = MAX(t1, t2) - objbounds_y;

	fld->x = objProp.x;
	fld->y = objProp.y;
	fld->wd = objProp.wd;
	fld->ht = objProp.ht;
	strcpy(fld->col, objProp.color[0]);
}

void DragAreaToProp(IEDragArea *dr, IEWindow *wind)
{
	curObj = dr;
	curWindNum = DRAGAREA_W;

	strcpy(objProp.name, dr->name);
	objbounds_x = objProp.x = dr->x;
	objbounds_y = objProp.y = dr->y;
	objbounds_wd = objProp.wd = dr->wd;
	objbounds_ht = objProp.ht = dr->ht;
	strcpy(objProp.color[0], dr->col);
	strcpy(objProp.funcName, dr->funcname);
	
	/*objProp.active = dr->active;*/
	objProp.bool1 = dr->active;
	
	objProp.objWind = wind;
	objbounds_x = dr->x;
	objbounds_y = dr->y;
	objbounds_wd = dr->wd;
	objbounds_ht = dr->ht;
}

void PropToDragArea()
{
	IEDragArea *dr = (IEDragArea *) curObj;
	int t1, t2;
	
	strcpy(dr->name, objProp.name);

	objbounds_x = MIN(dr->x, objProp.x);
	objbounds_y = MIN(dr->y, objProp.y);
	
	t1 = dr->x+dr->wd;
	t2 = objProp.x+objProp.wd;
	objbounds_wd = MAX(t1, t2) - objbounds_x;
	t1 = dr->y+dr->ht;
	t2 = objProp.y+objProp.ht;
	objbounds_ht = MAX(t1, t2) - objbounds_y;

	dr->x = objProp.x;
	dr->y = objProp.y;
	dr->wd = objProp.wd;
	dr->ht = objProp.ht;
	strcpy(dr->col, objProp.color[0]);
	strcpy(dr->funcname, objProp.funcName);
	/*dr->active = objProp.active;*/
	dr->active = objProp.bool1;
}

void GraphicToProp(IEGraphic *gr, IEWindow *wind)
{
	int nCols;
	int wflags;
	curObj = gr;

	objProp.objWind = wind;
	objbounds_x = objProp.x = gr->gr.x;
	objbounds_y = objProp.y = gr->gr.y;
	objbounds_wd = objProp.wd = gr->gr.wd;
	objbounds_ht = objProp.ht = gr->gr.ht;
	
	switch (gr->gr.type) {
	
		case GR_LINE:
		case GR_WLINE:
			curWindNum = LINE_W;
			objProp.bool1 = (gr->gr.type == GR_WLINE);
			objbounds_x = MIN(gr->gr.x, gr->gr.wd);
			objbounds_wd = abs(gr->gr.x - gr->gr.wd);
			objbounds_y = MIN(gr->gr.y, gr->gr.ht);
			objbounds_ht = abs(gr->gr.y - gr->gr.ht);
			break;
		
		case GR_RECT:
		case GR_WRECT:
			curWindNum = RECT_W;
			objProp.bool1 = (gr->gr.type == GR_WRECT);
			objProp.bool2 = (gr->gr.attrib == FILL);
			break;
			
		case GR_LABEL:
		case GR_LABEL_SMALL:
			curWindNum = LABEL_W;
			objProp.bool1 = (gr->gr.type == GR_LABEL);
			objProp.y += gr->gr.ht;
			break;
		}

	strcpy(objProp.color[0], gr->grcol);

	nCols = objParams[curWindNum].numColors;
	if (nCols >= 2)
		strcpy(objProp.color[1], gr->grshadowcol);

	wflags = objParams[curWindNum].windFlags;
	if ((wflags) & (HASTEXT))
		strcpy(objProp.text, gr->gr.text); 
}

void PropToGraphic()
{
	IEGraphic *gr = (IEGraphic *) curObj;
	int wflags = objParams[curWindNum].windFlags;
	int nCols = objParams[curWindNum].numColors;
	int t1, t2;
	int oldx, oldy, oldwd, oldht;
	
	if ((gr->gr.type == GR_LINE) || (gr->gr.type == GR_WLINE)) {
		objbounds_x = MIN(MIN(gr->gr.x, objProp.x),MIN(gr->gr.wd, objProp.wd));
		objbounds_y = MIN(MIN(gr->gr.y, objProp.y),MIN(gr->gr.ht, objProp.ht));
		objbounds_wd = MAX(abs(gr->gr.x - gr->gr.wd), abs(objProp.x - objProp.wd));
		objbounds_ht = MAX(abs(gr->gr.y - gr->gr.ht), abs(objProp.y - objProp.ht));
		objbounds_wd += objProp.bool1? THICK : THIN;
		objbounds_ht += objProp.bool1? THICK : THIN;
		}
	else if ((gr->gr.type == GR_RECT) || (gr->gr.type == GR_WRECT)) {
		objbounds_x = MIN(gr->gr.x, objProp.x);
		objbounds_y = MIN(gr->gr.y, objProp.y);
		
		t1 = gr->gr.x+gr->gr.wd;
		t2 = objProp.x+objProp.wd;
		objbounds_wd = MAX(t1, t2) - objbounds_x;
		
		t1 = gr->gr.y+gr->gr.ht;
		t2 = objProp.y+objProp.ht;
		objbounds_ht = MAX(t1, t2) - objbounds_y;
		}
	
	strcpy(gr->grcol, objProp.color[0]);
	if (nCols >= 2)
		strcpy(gr->grshadowcol, objProp.color[1]);
	if ((wflags) & (HASTEXT))
		strcpy(gr->gr.text, objProp.text); 
	
	switch (gr->gr.type) {
				
		case GR_LINE:
		case GR_WLINE:
			gr->gr.type = (objProp.bool1? GR_WLINE : GR_LINE);
			gr->gr.x = objProp.x;
			gr->gr.y = objProp.y;
			gr->gr.wd = objProp.wd;
			gr->gr.ht = objProp.ht;
		break;
		
		case GR_RECT:
		case GR_WRECT:
			gr->gr.type = (objProp.bool1? GR_WRECT : GR_RECT);
			gr->gr.attrib = (objProp.bool2? FILL : HOLLOW);
			gr->gr.x = objProp.x;
			gr->gr.y = objProp.y;
			gr->gr.wd = objProp.wd;
			gr->gr.ht = objProp.ht;
			break;
			
		case GR_LABEL:
		case GR_LABEL_SMALL:
			oldx = gr->gr.x;
			oldy = gr->gr.y;
			oldwd = gr->gr.wd;
			oldht = gr->gr.ht;
			
			gr->gr.type = (objProp.bool1? GR_LABEL : GR_LABEL_SMALL);
			set_label_dims(gr, objProp.x, objProp.y);

			objbounds_x = MIN(gr->gr.x, oldx);
			objbounds_y = MIN(gr->gr.y, oldy);
			
			t1 = gr->gr.x+gr->gr.wd;
			t2 = oldx + oldwd;
			objbounds_wd = MAX(t1, t2) - objbounds_x;
			
			t1 = gr->gr.y + gr->gr.ht;
			t2 = oldy + oldht;
			objbounds_ht = MAX(t1, t2) - objbounds_y+1;
			break;
		}	
}

void PopupToProp(IEPopup *pm)
{
	curObj = pm;
	objProp.objWind = NULL;

	strcpy(objProp.name, pm->title);
	objProp.x = pm->x;
	objProp.y = pm->y;
	objProp.wd = pm->wd;
	objProp.ht = pm->ht;
	objProp.bool1 = pm->checklist;
	strcpy(objProp.color[0], pm->efg);
	strcpy(objProp.color[1], pm->dfg);
	strcpy(objProp.color[2], pm->bg);
	strcpy(objProp.color[3], pm->bcolor);
	strcpy(objProp.funcName, pm->function);
	
	curWindNum = POPUP_W;

	objbounds_x = pm->x;
	objbounds_y = pm->y;
	objbounds_wd = pm->wd;
	objbounds_ht = pm->ht;
}

void PropToPopup()
{
	IEPopup *pm = (IEPopup *) curObj;
	int t1, t2;
	
	if (curWindNum == POPUP_W)
		RemovePlabStuff();

	objbounds_x = MIN(pm->x, objProp.x);
	objbounds_y = MIN(pm->y, objProp.y);
	
	t1 = pm->x+pm->wd;
	t2 = objProp.x+objProp.wd;
	objbounds_wd = MAX(t1, t2) - objbounds_x;
	
	t1 = pm->y+pm->ht;
	t2 = objProp.y+objProp.ht;
	objbounds_ht = MAX(t1, t2) - objbounds_y;
			
	pm->x = objProp.x;
	pm->y = objProp.y;
	objProp.wd = pm->wd;
	objProp.ht = pm->ht;
	objProp.bool1 = pm->checklist;
	strcpy(pm->title, objProp.name);
	strcpy(pm->efg, objProp.color[0]);
	strcpy(pm->dfg, objProp.color[1]);
	strcpy(pm->bg, objProp.color[2]);
	strcpy(pm->bcolor, objProp.color[3]);
	strcpy(pm->function, objProp.funcName);
}

void DelCurrentObject()
{
	switch(curWindNum) {
		
		case FIELD_W:
			DeleteField(objProp.objWind, (IEField *) curObj);
			break;
			
		case DRAGAREA_W:
			DeleteDragArea(objProp.objWind, (IEDragArea *) curObj);
			break;
			
		case BUTTON_W:
			DeleteButton(objProp.objWind, (IEButton *) curObj);
			break;
			
		case RECT_W:
		case LINE_W:
		case LABEL_W:
			DeleteGraphic(objProp.objWind, (IEGraphic *) curObj);
			break;
			
		case POPUP_W:
			DeletePopup((IEPopup *) curObj);
			break;
		}
}	

void DrawPropWind(boolean delOnCancel)
{
	int wflags = objParams[curWindNum].windFlags;
	
	if (curWindNum == NUM_PROP_WINDS)
		return;
	
	colorList = funcList = -1;
	
	cancelDel = delOnCancel;
	
	if ((wflags) & (BOOL1))
		GSsetbuttontext(objParams[curWindNum].windName, "ToggleBool1",
			objProp.bool1? objParams[curWindNum].Bool1True : objParams[curWindNum].Bool1False);
	
	if ((wflags) & (BOOL2))
		GSsetbuttontext(objParams[curWindNum].windName, "ToggleBool2",
			objProp.bool2? objParams[curWindNum].Bool2True : objParams[curWindNum].Bool2False);
	
	objwind = TRUE;	
	GSdrawnamedwindow(objParams[curWindNum].windName, BSTORE, POPUP);

	if ((wflags) & (NAME))
		name = NewEditRec(objProp.name, BIGSTRINGLEN, 1, SMALL, 
			objParams[curWindNum].windName, "PropName", c_identifier_filter, NULL);

	if ((wflags) & (HASTEXT))
		text = NewEditRec(objProp.text, BIGSTRINGLEN, 1, SMALL, 
			objParams[curWindNum].windName, "PropText", NULL, NULL);

	if ((wflags) & (FUNCREC))
		func = NewEditRec(objProp.funcName, BIGSTRINGLEN, 1, SMALL,
			objParams[curWindNum].windName, "PropFunc", c_identifier_filter, NULL);

	if ((wflags) & (COLORLIST))
		ObjSetupColorList();
	
	if ((wflags) & (FUNCLIST))
		ObjSetupFuncList();
		
	if ((wflags) & (LABPOS)) {
		SetLabPosButtonText();
		DisplayLabPosButton();
		}
		
	if ((wflags) & (ACTIVE)) {
		SetActiveButtonText();
		DisplayActiveButton();
		}

	if (curWindNum == POPUP_W)
		DrawPlabStuff((IEPopup *) curObj);

	DisableColorButtons();
	ShowPropWind();
	GScharfunc("process_key");
}

static void ObjSetupColorList()
{
	int x, y, wd, dragHt;
	int wflags = objParams[curWindNum].windFlags;

	for (numListColors = 0; Colorlist[numListColors] != NULL; numListColors++);
	for (numColors = numListColors; (Alias[numColors-numListColors] != NULL)
					&& (*Alias[numColors-numListColors] != '\0'); numColors++);

	GSgetdragareaspecs(objParams[curWindNum].windName, 
						"propColPick", &x, &y, &wd, &dragHt, NULL, NULL);
	curColor = NO_ITEM;
	
	GSsclistsetup(objParams[curWindNum].windName, 
					"propColPick",
					1,
					"propPickLabel",
					0, 0,
					JUST_LEFT,
					objLabelFunc,
					objTextFunc, 
					objInFunc,
					objOutFunc,
					objClickFunc,
					"propColScroll",
					GSconvertcolor(THUMBCOL),
					GSconvertcolor(THUMBTL),
					GSconvertcolor(THUMBBR),
					numColors+( ((wflags)&&(CLIST_HAS_NONE))? 1 : 0), 0, 
					NULL,
					"propColUp", "propColDown", 
					&colorList, &colorScroll);
							
					
	ShadowRect(x, y, wd, dragHt, THIN, Concave);
	GSdrawsclist(colorScroll);
}

static void ObjSetupFuncList() /* popups only, right now */
{
	int x, y, wd, dragHt;
	IEPopup *p = iestate.popuplisthead;
	
	for(numFuncs=0; p != NULL; p = p->nextpopup, numFuncs++)
		;
		
	GSgetdragareaspecs(objParams[curWindNum].windName, 
						"propFuncPick", &x, &y, &wd, &dragHt, NULL, NULL);
	curFunc = NO_ITEM;
	
	GSsclistsetup(objParams[curWindNum].windName, 
					"propFuncPick",
					1,
					"propPickLabel",
					0, 0,
					JUST_LEFT,
					fnLabelFunc,
					objTextFunc, 
					fnInFunc,
					fnOutFunc,
					fnClickFunc,
					"propFuncScroll",
					GSconvertcolor(THUMBCOL),
					GSconvertcolor(THUMBTL),
					GSconvertcolor(THUMBBR),
					numFuncs, 0,
					NULL,
					"propFuncUp", "propFuncDown",
					&funcList, &funcScroll);
					
	ShadowRect(x, y, wd, dragHt, THIN, Concave);
	GSdrawsclist(colorScroll);
}

void RedrawPropWind()
{
	GSredrawnamedwindow(objParams[curWindNum].windName);
	ShowPropWind();
	if (curWindNum == POPUP_W)
		RedrawPlabStuff();
}

void RemovePropWind(boolean userOK)
{
	int wflags = objParams[curWindNum].windFlags;
	int clipx, clipy, clipwd, clipht;
	
	objwind = FALSE;

	if (curWindNum == POPUP_W)
		DismantlePlabStuff();

	if ((wflags) & (COLORLIST))
		GSdismantlesclist(colorScroll);
	if ((wflags) & (FUNCLIST))
		GSdismantlesclist(funcScroll);
	if ((wflags) & (HASTEXT))
		DisposEditRec(text);
	if ((wflags) & (NAME))
		DisposEditRec(name);
	if ((wflags) & (FUNCREC))
		DisposEditRec(func);
		
	GSremovenamedwindow(objParams[curWindNum].windName);
	GScharfunc("ikey");

	if (cancelDel && !userOK) {
		DelCurrentObject();
		}
		
	if (userOK || (cancelDel && !userOK)) {
		/* redraw interface to reflect changes */
		/*int x, y;*/
		IEWindow *w;
		
		if (curWindNum != POPUP_W) {
			w = objProp.objWind;
			objbounds_x += w->x;
			objbounds_y += w->y;
			}
		else
			w = iestate.winlisthead;
						
		if (twindVisible)
			GSremovenamedwindow("toolw");
					
		PushClip(objbounds_x, objbounds_y, objbounds_wd, objbounds_ht);
		DrawFromWind(w, objbounds_x, objbounds_y, objbounds_wd, objbounds_ht);
		PopClip();
		
		if (twindVisible)
			GSdrawnamedwindow("toolw", BSTORE, NO_POPUP);
		}
	
	curWindNum = NUM_PROP_WINDS;
}

void ShowPropWind()
{
	int wflags = objParams[curWindNum].windFlags;
	int x, y, wd, ht;
	
	GSsetcurrentnamedwindow(objParams[curWindNum].windName);
	
	DrawNumField("PropXDisp", objProp.x);
	DrawNumField("PropYDisp", objProp.y);
	
	if ((wflags) & (NAME)) {
		UpdateEditRec(name);
		}
	
	if ((wflags) & (HASTEXT)) {
		UpdateEditRec(text);
		}
			
	if ((wflags) & (FUNCREC)) {
		UpdateEditRec(func);
		}
			
	if ((wflags) & (WD_HT)) {
		DrawNumField("PropWdDisp", objProp.wd);
		DrawNumField("PropHtDisp", objProp.ht);
		}
		
	if ((wflags) & (ACTIVE)) {
/*		SetActiveButtonText();
		DisplayActiveButton();*/
		}
		
	if ((wflags) & (COLORLIST)) {
		GSdrawsclist(colorScroll);
		GSgetdragareaspecs(objParams[curWindNum].windName, 
				"propColPick", &x, &y, &wd, &ht, NULL, NULL);
		ShadowRect(x, y, wd, ht, THIN, Concave);
		}
	
	if ((wflags) & (FUNCLIST)) {
		GSdrawsclist(funcScroll);
		GSgetdragareaspecs(objParams[curWindNum].windName, 
				"propFuncPick", &x, &y, &wd, &ht, NULL, NULL);
		ShadowRect(x, y, wd, ht, THIN, Concave);
		}
					
/*	if ((wflags) & (COLOR1))
		ShowColorButtonFields(F_COLOR1);
	if ((wflags) & (COLOR2))
		ShowColorButtonFields(F_COLOR2);
	if ((wflags) & (COLOR3))
		ShowColorButtonFields(F_COLOR3);*/
	
	ShowColorButtonFields(F_ALL);
		
}

void DisplayActiveButton()
{
	GSenablebutton(objParams[curWindNum].windName, "SetPropActive", TRUE);
}

void SetActiveButtonText()
{
	switch(objProp.active) {
		case INVISIBLE:
			GSsetbuttontext(objParams[curWindNum].windName, 
									"SetPropActive", "Invisible");
			break;
			
		case DISABLED:
			GSsetbuttontext(objParams[curWindNum].windName, 
									"SetPropActive", "Inactive");
			break;
			
		case  ENABLED:
			GSsetbuttontext(objParams[curWindNum].windName, 
									"SetPropActive", "Active");
			break;
		}
}

void SetLabPosButtonText()
{
	switch(objProp.labpos) {
		case LEFT:
			GSsetbuttontext(objParams[curWindNum].windName,
									"SetPropLabPos", "Left");
			break;
			
		case CENTER:
			GSsetbuttontext(objParams[curWindNum].windName,
									"SetPropLabPos", "Center");
			break;
			
		case RIGHT:
			GSsetbuttontext(objParams[curWindNum].windName,
									"SetPropLabPos", "Right");
			break;
			
		}
}

void DisplayLabPosButton()
{
	GSenablebutton(objParams[curWindNum].windName, "SetPropLabPos", TRUE);
}

static void EnableColorButtons()
{
	int nCols = objParams[curWindNum].numColors;
	char *wname = objParams[curWindNum].windName;
	
	if (nCols >= 2) {
		GSenablebutton(wname, "SetPropCol2", TRUE);
		ShowColorButtonFields(F_COLOR2);
		}
	
	if (nCols >= 3) {
		GSenablebutton(wname, "SetPropCol3", TRUE);
		ShowColorButtonFields(F_COLOR3);
		}
	
	if (nCols >= 4) {
		GSenablebutton(wname, "SetPropCol4", TRUE);
		ShowColorButtonFields(F_COLOR4);
		}
}

static void EnableColor1()
{
	char *wname = objParams[curWindNum].windName;
	int nCols = objParams[curWindNum].numColors;
	if (nCols >= 1) {
		GSenablebutton(wname, "SetPropCol1", TRUE);
		ShowColorButtonFields(F_COLOR1);
		}
}

static void DisableColor1()
{
	char *wname = objParams[curWindNum].windName;
	int nCols = objParams[curWindNum].numColors;
	if (nCols >= 1) {
		GSdisablebutton(wname, "SetPropCol1", TRUE);
		ShowColorButtonFields(F_COLOR1);
		}
}

static void DisableColorButtons()
{
	int nCols = objParams[curWindNum].numColors;
	char *wname = objParams[curWindNum].windName;
	
	if (nCols >= 1)
		GSdisablebutton(wname, "SetPropCol1", TRUE);
	
	if (nCols >= 2)
		GSdisablebutton(wname, "SetPropCol2", TRUE);
	
	if (nCols >= 3)
		GSdisablebutton(wname, "SetPropCol3", TRUE);

	if (nCols >= 4)
		GSdisablebutton(wname, "SetPropCol4", TRUE);

/*	ShowColorButtonFields(F_ALL);*/
}

#define FIELD_LEN 70

void ShowColorButtonFields(int which)
{
	int x, y, wd, ht;
/*	int wflags = objParams[curWindNum].windFlags;*/
	int nCols = objParams[curWindNum].numColors;
	char *names[F_ALL] = {"SetPropCol1",
						  "SetPropCol2",
						  "SetPropCol3",
						  "SetPropCol4"};
	int i;

	for (i=0; i<F_ALL; i++)
		if ( ((which==F_ALL)||(which==i)) && (i < nCols) ){
			/*GSgetfieldrect(names[i], &x, &y, &wd, &ht, &col);*/
			GSgetbuttonspecs(objParams[curWindNum].windName, 
								names[i], &x, &y, &wd, &ht, NULL);
			
			ht /= 2;
			y += ht/2;
			x = (x+wd) - ht - FIELD_LEN;
			wd = FIELD_LEN;
			
			GSsetcolor(GSconvertcolor("white"));
			GSwfillrect(x, y, wd, ht);
			ShadowRect(x, y, wd, ht, 1, Concave);
			GSsetcolor(blackCol);
			GSwwritemsg(x, y+SMALL, objProp.color[i]);
			}
}

boolean set_prop_x(char *bName)
{
	objProp.x += GET_INC(*bName);
	DrawNumField("PropXDisp", objProp.x);
}

boolean set_prop_y(char *bName)
{
	objProp.y += GET_INC(*bName);
	DrawNumField("PropYDisp", objProp.y);
}

boolean set_prop_wd(char *bName)
{
	if (objParams[curWindNum].windFlags & (WD_HT)) {
		objProp.wd += GET_INC(*bName);
		DrawNumField("PropWdDisp", objProp.wd);
		}
}

boolean set_prop_ht(char *bName)
{
	if ((objParams[curWindNum].windFlags) & (WD_HT)) {
		objProp.ht += GET_INC(*bName);
		DrawNumField("PropHtDisp", objProp.ht);
		}
}

boolean set_prop_col1()
{
	strcpy(objProp.color[0], GetColor(curColor));
	ShowColorButtonFields(F_COLOR1);
}

boolean set_prop_col2()
{
	strcpy(objProp.color[1], GetColor(curColor));
	ShowColorButtonFields(F_COLOR2);
}

boolean set_prop_col3()
{
	strcpy(objProp.color[2], GetColor(curColor));
	ShowColorButtonFields(F_COLOR3);
}

boolean set_prop_col4()
{
	strcpy(objProp.color[3], GetColor(curColor));
	ShowColorButtonFields(F_COLOR4);
}

boolean prop_col_pick(int x, int y, DragStatusType status)
{
	GShandlelist(colorList, x, y, status);
	return TRUE;
}

boolean prop_col_scroll(int x, int y, DragStatusType status)
{
	GShandlesclist(colorScroll, x, y, status);
	return TRUE;
}

boolean prop_col_btn(char *bName)
{
	int r = GET_INC(bName[7]);
	if (r > 0)
		GShandleupbutton(colorScroll);
	else
		GShandledownbutton(colorScroll);
}

boolean prop_func_pick(int x, int y, DragStatusType status)
{
	GShandlelist(funcList, x, y, status);
	return TRUE;
}

boolean prop_func_scroll(int x, int y, DragStatusType status)
{
	GShandlesclist(funcScroll, x, y, status);
	return TRUE;
}

boolean prop_func_btn(char *bName)
{
	int r = GET_INC(bName[8]);
	if (r > 0)
		GShandleupbutton(funcScroll);
	else
		GShandledownbutton(funcScroll);
}

boolean toggle_prop_bool1()
{
	objProp.bool1 = !objProp.bool1;
	
	GSsetbuttontext(objParams[curWindNum].windName, "ToggleBool1",
		objProp.bool1? objParams[curWindNum].Bool1True : objParams[curWindNum].Bool1False);
	
	GSenablebutton(objParams[curWindNum].windName, "ToggleBool1", TRUE);
}

boolean toggle_prop_bool2()
{
	objProp.bool2 = !objProp.bool2;
	
	GSsetbuttontext(objParams[curWindNum].windName, "ToggleBool2",
		objProp.bool2? objParams[curWindNum].Bool2True : objParams[curWindNum].Bool2False);
	
	GSenablebutton(objParams[curWindNum].windName, "ToggleBool2", TRUE);
}

boolean pname_prop(int x, int y, DragStatusType status)
{
	if (objParams[curWindNum].windFlags & (NAME))
		if (status == MOUSE_CLICK || status == DRAG_FINAL)
			SetCurrentEditRec(name);
}

boolean ptext_prop(int x, int y, DragStatusType status)
{
	if (objParams[curWindNum].windFlags & (HASTEXT))
		if (status == MOUSE_CLICK || status == DRAG_FINAL)
			SetCurrentEditRec(text);
}

boolean pfunc_prop(int x, int y, DragStatusType status)
{
	if (objParams[curWindNum].windFlags & (FUNCREC))
		if (status == MOUSE_CLICK || status == DRAG_FINAL)
			SetCurrentEditRec(func);
}

boolean set_prop_func()
{
}

boolean prop_ok()
{
	switch(curWindNum) {
		
		case FIELD_W:
			PropToField();
			break;
			
		case DRAGAREA_W:
			PropToDragArea();
			break;
			
		case BUTTON_W:
			PropToButton();
			break;
			
		case RECT_W:
		case LINE_W:
		case LABEL_W:
			PropToGraphic();
			break;
			
		case POPUP_W:
			PropToPopup();
			break;
		}
	
	RemovePropWind(TRUE);
}

boolean prop_cancel()
{
	RemovePropWind(FALSE);
}

boolean prop_delete() /* +++ may want to warn before doing this */
{
	if (curWindNum == LABEL_W)  /* hack!! */
		objProp.y -= objProp.ht;
	DelCurrentObject();
	RemovePropWind(TRUE);
}

boolean set_prop_active(char *item_name)
{
	int inum = GSgetmenuitemnum("SetPropActive", item_name);
	
	switch(inum) {
	
		case 0:
			objProp.active = INVISIBLE;
			break;
			
		case 1:
			objProp.active = DISABLED;
			break;
			
		case 2:
			objProp.active = ENABLED;
			break;
		}
		
	if (objParams[curWindNum].windFlags & (ACTIVE)) {
		SetActiveButtonText();
		DisplayActiveButton();
		}
}

boolean set_prop_labpos(char *item_name)
{
	int inum = GSgetmenuitemnum("SetPropLabpos", item_name);
	
	switch(inum) {
		
		case 0:
			objProp.labpos = LEFT;
			break;
			
		case 1:
			objProp.labpos = CENTER;
			break;
			
		case 2:
			objProp.labpos = RIGHT;
			break;
		}
	
	if (objParams[curWindNum].windFlags & (ACTIVE)) {
		SetLabPosButtonText();
		DisplayLabPosButton();
		}
}

static char *GetColor(int curColor)
{
	if (curColor < numListColors)
		return Colorlist[curColor];
	else if (curColor < numColors)
		return Alias[curColor - numListColors];
	else if (curColor == numColors)
		return "NONE";
	else
		return "error!";
}
