#include "geosim.h"
#include "gspriv.h"
#include "ietypes.h"
#include "windfunc.h"
#include "windprop.h"
#include "editrec.h"
/*#include "rootw.h"*/
#include "iemisc.h"
#include "ifw.h"
#include <string.h>

#ifndef MSDOS
#include <stdlib.h>
#endif

#define MIN_BORDER 0
#define MAX_BORDER 10

#define F_ALL 3
#define F_FG 0
#define F_BG 1
#define F_BDR 2

extern boolean windPropPop;
extern GSColor blackCol;
/*extern char *IEColorlist[], *IEAlias[];
Shouldn't touch these here, right?
*/
extern char *Colorlist[], *Alias[];
extern IEIconfig iestate;

static IEWindow propWind, *realWind;
static EditRec *nameEdit;
static int numColors, numListColors, curColor;
static GSlistHandle wColorList;
static GSsclistHandle wColorScroll;
static boolean cancelDel;
/*static BigString tempName;*/

/* Internal protos */
void DrawNumField(char *fName, int val);
static void InitButtons();
void DupWind(IEWindow *src, IEWindow *dest);
void ShowBStoreField();
static void PSetupColorList();
void ShowWColorFields(int);
static void EnableColorButtons();
static void DisableColorButtons();
/**/

static void pTextFunc(GSlistHandle list, int slot)
{
	if (slot == curColor)
		GSsetcolor(GSconvertcolor("dgray"));
	else
		GSsetcolor(blackCol);
	GSsetfont(SMALL);
}

static char *pLabelFunc(int slot)
{
	if (slot < numListColors)
		return Colorlist[slot];
	else if (slot < numColors)
		return Alias[slot-numListColors];
	else
		return "";
}

static void pInFunc()
{
}

static void pOutFunc()
{
}

static void pClickFunc(int slot)
{
	int old = curColor;
	
	curColor = slot;
	
	if (old == -1)
		EnableColorButtons();
	
	GSdrawlistitem(wColorList, New, old);
	GSdrawlistitem(wColorList, New, curColor);
}

void DupWind(IEWindow *src, IEWindow *dest)
{
	memcpy(dest, src, sizeof(IEWindow));
}

static void InitButtons()
{
	if (propWind.bordersize > MIN_BORDER)
		GSenablebutton("windprop", "IncWindBSize", TRUE);
	else
		GSdisablebutton("windprop", "DecWindBSize", TRUE);
}

void DrawWPropWind(IEWindow *wind, boolean deleteOnCancel)
{
	realWind = wind;
	DupWind(realWind, &propWind);
	windPropPop = TRUE;
	cancelDel = deleteOnCancel;
	
	GSsetbuttontext("windprop", "ToggleWindActive", 
					propWind.active? "Active" : "Inactive");
	GSdrawnamedwindow("windprop", POPUP, BSTORE);

	GScharfunc("process_key");
	
	nameEdit = NewEditRec(propWind.name, BIGSTRINGLEN, 1, SMALL,
			"windprop", "WindName", c_identifier_filter, NULL);
	
	SetCurrentEditRec(nameEdit);
	UpdateEditRec(nameEdit);
	DrawNumField("WindXDisp", propWind.x);
	DrawNumField("WindYDisp", propWind.y);
	DrawNumField("WindWdDisp", propWind.wd);
	DrawNumField("WindHtDisp", propWind.ht);
	DrawNumField("WindBSizeDisp", propWind.bordersize);
	
	InitButtons();
	ShowBStoreField();
	PSetupColorList();
	/*ShowWColorFields(F_ALL);*/
	DisableColorButtons();
}

void RedrawWPropWind()
{
	int x, y, wd, ht;
	GSredrawnamedwindow("windprop");
	GSsetactivenamedwindow("windprop");
	UpdateEditRec(nameEdit);
	DrawNumField("WindXDisp", propWind.x);
	DrawNumField("WindYDisp", propWind.y);
	DrawNumField("WindWdDisp", propWind.wd);
	DrawNumField("WindHtDisp", propWind.ht);
	DrawNumField("WindBSizeDisp", propWind.bordersize);
	ShowBStoreField();
	GSgetdragareaspecs("windprop", "wColPick", &x, &y, 
								&wd, &ht, NULL, NULL);
	ShadowRect(x, y, wd, ht, THIN, Concave);
	GSdrawsclist(wColorScroll);
	ShowWColorFields(F_ALL);
}

void RemoveWPropWind()
{
	GScharfunc("ikey");

	windPropPop = FALSE;
	DisposEditRec(nameEdit);
	GSremovenamedwindow("windprop");
	GSdismantlesclist(wColorScroll);
	DrawInterface(TRUE);
}

static void PSetupColorList()
{
	int x, y, wd, ht;
	curColor = NO_ITEM;
	
	for (numListColors = 0; 
		(Colorlist[numListColors] != NULL) &&
			(*Colorlist[numListColors] != '\0'); numListColors++);
	
	for(numColors = numListColors; 
		(Alias[numColors-numListColors] != NULL) &&
			(*Alias[numColors-numListColors] != '\0'); numColors++);
			
	GSsclistsetup("windprop", "wColPick",
					1,
					"smpicklabel",
					0, 0,
					JUST_LEFT,
					pLabelFunc,
					pTextFunc,
					pInFunc,
					pOutFunc,
					pClickFunc,
					"wColScroll",
					GSconvertcolor(THUMBCOL),
					GSconvertcolor(THUMBTL),
					GSconvertcolor(THUMBBR),
					numColors, 0, 
					NULL,
					"wColUp", "wColDown",
					&wColorList, &wColorScroll					
					);

	GSgetdragareaspecs("windprop", "wColPick", &x, &y, 
								&wd, &ht, NULL, NULL);
	ShadowRect(x, y, wd, ht, THIN, Concave);
	GSdrawsclist(wColorScroll);
}

void ShowWColorFields(int which)
{
	int x, y, wd, ht;
	int i;
	GSColor col;
	char *names[F_ALL][2] = {NULL, "WindFGDisp",
							NULL, "WindBGDisp",
							NULL, "WindBdrDisp"};
	names[F_FG][0] = propWind.fgcol;
	names[F_BG][0] = propWind.bgcol;
	names[F_BDR][0] = propWind.bdrcol;
	
	for (i=0; i<F_ALL; i++)
		if ((which == F_ALL)||(which == i)) {
			GSgetfieldrect(names[i][1], &x, &y, &wd, &ht, &col);
			GSsetcolor(GSconvertcolor("white"));
			GSwfillrect(x, y, wd, ht);
			GSsetlinesize(1);
			GSframefield("windprop", names[i][1], 
				GSconvertcolor("black"), GSconvertcolor("dgray"));
			GSsetcolor(col);
			GSwwritemsg(x, y+SMALL, names[i][0]);
			}
}

static void EnableColorButtons()
{
	GSenablebutton("windprop", "SetWindFGCol", TRUE);
	GSenablebutton("windprop", "SetWindBGCol", TRUE);
	GSenablebutton("windprop", "SetWindBdrCol", TRUE);
	
	ShowWColorFields(F_ALL);
}

static void DisableColorButtons()
{
	GSdisablebutton("windprop", "SetWindFGCol", TRUE);
	GSdisablebutton("windprop", "SetWindBGCol", TRUE);
	GSdisablebutton("windprop", "SetWindBdrCol", TRUE);
	
	ShowWColorFields(F_ALL);
}

void ShowBStoreField()
{
	int x, y, wd, ht;
	GSColor col;
	char temp[10];
	
	GSgetfieldrect("WindBStoreDisp", &x, &y, &wd, &ht, &col);
	GSsetcolor(GSconvertcolor("white"));
	GSwfillrect(x, y, wd, ht);
	GSsetlinesize(1);
	GSframefield("windprop", "WindBStoreDisp", 
						blackCol, GSconvertcolor("dgray"));
	GSsetcolor(col);
	sprintf(temp, "%d", propWind.bstore_to_use);
	GSwwritemsg(x, y+SMALL, temp);
}

boolean set_wind_bstore(char *itemName)
{
	propWind.bstore_to_use = itemName[3] - '0';
	ShowBStoreField();
}

boolean wname_prop(int x, int y, DragStatusType t)
{
	SetCurrentEditRec(nameEdit);
}

boolean set_wind_x(char *bName)
{
	/* buttons that start with 'd' are "decrement" buttons */
	propWind.x += GET_INC(*bName);
	DrawNumField("WindXDisp", propWind.x);
}

boolean set_wind_y(char *bName)
{
	propWind.y += GET_INC(*bName);
	DrawNumField("WindYDisp", propWind.y);
}

boolean set_wind_wd(char *bName)
{
	propWind.wd += GET_INC(*bName);
	DrawNumField("WindWdDisp", propWind.wd);
}

boolean set_wind_ht(char *bName)
{
	propWind.ht += GET_INC(*bName);
	DrawNumField("WindHtDisp", propWind.ht);
}

boolean set_wind_bsize(char *bName)
{
	propWind.bordersize += GET_INC(*bName);
	
	if (propWind.bordersize==MAX_BORDER)
		GSdisablebutton("windprop", "IncWindBSize", TRUE);
	else if (propWind.bordersize==MAX_BORDER-1)
		GSenablebutton("windprop", "IncWindBSize", TRUE);
			
	if (propWind.bordersize==MIN_BORDER)
		GSdisablebutton("windprop", "DecWindBSize", TRUE);
	else if (propWind.bordersize==MIN_BORDER+1)
		GSenablebutton("windprop", "DecWindBSize", TRUE);
			
	DrawNumField("WindBSizeDisp", propWind.bordersize);
}

boolean toggle_wind_active()
{
	propWind.active = !propWind.active;
	
	GSsetbuttontext("windprop", "ToggleWindActive", 
					propWind.active? "Active" : "Inactive");	
	GSenablebutton("windprop", "ToggleWindActive", TRUE);
}

boolean set_wind_fg()
{
	char *srcStr;
	
	if (curColor == NO_ITEM)
		return TRUE;
	
	if (curColor < numListColors)
		srcStr = Colorlist[curColor];
	else if (curColor < numColors)
		srcStr = Alias[curColor - numListColors];
	else
		return TRUE;
		
	strcpy(propWind.fgcol, srcStr);
	ShowWColorFields(F_FG);
}

boolean set_wind_bg()
{
	char *srcStr;
	
	if (curColor == NO_ITEM)
		return TRUE;
	
	if (curColor < numListColors)
		srcStr = Colorlist[curColor];
	else if (curColor < numColors)
		srcStr = Alias[curColor - numListColors];
	else
		return TRUE;
		
	strcpy(propWind.bgcol, srcStr);
	ShowWColorFields(F_BG);
}

boolean set_wind_bdr()
{
	char *srcStr;
	
	if (curColor == NO_ITEM)
		return TRUE;
	
	if (curColor < numListColors)
		srcStr = Colorlist[curColor];
	else if (curColor < numColors)
		srcStr = Alias[curColor - numListColors];
	else
		return TRUE;
		
	strcpy(propWind.bdrcol, srcStr);
	ShowWColorFields(F_BDR);
}

boolean w_col_pick(int x, int y, DragStatusType status)
{
	GShandlelist(wColorList, x, y, status);
}

boolean w_col_scroll(int x, int y, DragStatusType status)
{
	GShandlesclist(wColorScroll, x, y, status);
}

boolean w_col_btn(char *name)
{
	if (name[4] == 'U')  /* hack alert */
		GShandleupbutton(wColorScroll);
	else if (name[4] == 'D')
		GShandledownbutton(wColorScroll);
	else
		GSbell();
}

boolean w_ok()
{
	DupWind(&propWind, realWind);
	RemoveWPropWind();
}

boolean w_cancel()
{
	if (cancelDel)
		DeleteWindow(realWind);
	RemoveWPropWind();
}

boolean w_delete()
{
	DeleteWindow(realWind);
	RemoveWPropWind();
}
