/***********************************************************************
Mental Maps module interface functions initiated by the GeoSim
interface handler.
***********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "geosim.h"
#include "mmincl.h"
#include "triang.h"

/*ColorIndexType fastrow;  */
GSColor fastrow;
BigString      Helpfile;
BigString      CurrWin;
extern Point2  Citypoints[72];
int            StepCount;
int            Xorx, Xory, Xorwd, Xorht;
BigString      LabelBeingMoved;
BigString      CityBeingMoved;
boolean        Got_city;
boolean        Got_label;

void    initiate_move     (int, int);
void    finish_move       (void);
void    continue_move     (int, int);
void    compute_xorbox    (int, int, int, int, int, int);
void    highlightfield    (BigString, BigString, GSColor, GSColor, GSColor);


/*------------------------------------------------------------------------------
  Cancel quit.  Erase quit window and return to program.
------------------------------------------------------------------------------*/
boolean cancelquit(void)
{
  GSremovenamedwindow("quit");
  QuitWinActive = FALSE;          
  if (!Use_bstore)
  {
    redraw_all();
  }
  GSsetcurrentnamedwindow(CurrWin);
  GSdisplayhelpmessage (PREV_MSG);
  return TRUE;
}

/*------------------------------------------------------------------------------
  Confirm quit.  Actually quit the program
------------------------------------------------------------------------------*/
boolean confirmquit(void)
{
  if (!LoginInProcess)  
    GSdisplayhelpmessage (27);
  GSgetcurrentwindowname(CurrWin);
  GSdrawnamedwindow("quit", Use_bstore, POPUP);
  QuitWinActive = TRUE;          
  return TRUE;
}

/*----------------------------------------------------------------------------
compute location of xorbox.
----------------------------------------------------------------------------*/
void compute_xorbox
 (int cx,
  int cy,
  int fwd,
  int fht,
  int xorwd,
  int xorht)
{
  /*--- cx, cy should be center of the box, but move it to fit in screen
        if necessary ---*/
  Xorwd = xorwd;
  Xorht = xorht;
  Xorx = cx - Xorwd/2;
  if (Xorx < 0)
    Xorx = 0;
  if ((Xorx + Xorwd) > fwd)
    Xorx = fwd - Xorwd;
  Xory = cy - Xorht/2;
  if (Xory < 0)
    Xory = 0;
  if ((Xory + Xorht) > fht)
    Xory = fht - Xorht;
}

/*----------------------------------------------------------------------
  Initiate "about questions" help window.
  GLOBALS: Helpfile, Infopath, NotInStepThru
----------------------------------------------------------------------*/
boolean aboutq(void)
{
  sprintf(Helpfile, "%s%s", Infopath, QUESTION_HLP);
  HelpWinActive = TRUE;
  return(GSdohelpscreens(Helpfile, FirstPage, Use_bstore, 10));
}


/*----------------------------------------------------------------------
  "Done" button in the alert box.
  GLOBALS:
----------------------------------------------------------------------*/
boolean alertok(void)
{
  GSremovenamedwindow("alert");
  DumpResultsActive = FALSE;
  AlertNumber = 0;
  if (!Use_bstore)
    redraw_all();
  GSdisplayhelpmessage(Savemsg); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/*----------------------------------------------------------------------
  Display help message when "Country" button is pressed.
  GLOBALS:
----------------------------------------------------------------------*/
boolean countryhelp(void)
{
  GSdisplayhelpmessage(3); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/*----------------------------------------------------------------------
  Handle an item on the country menu.
  GLOBALS: Savemsg, MapFile, WarpMapFile, CityFile, City
----------------------------------------------------------------------*/
boolean countryitem(char *itemname)
{ 
  /*--- show message before drawing map to overwrite old Savemsg which
        is redisplayed when countryitem is selected ---*/
  int itemnum = GSgetmenuitemnum("Country", itemname);
  if (Moving)
    GSdisplayhelpmessage(Savemsg = 26);
  else
    GSdisplayhelpmessage(Savemsg = 4);

  FinalMapActive = FALSE;

  GSdisabledragarea("Map", "Dragmap");
  Curr_country = itemnum;
  if (City[Curr_country].hi_avail == FALSE
   && City[Curr_country].low_avail == FALSE)
    GSdisablemenuitem("quest", GSgetmenuitemname("quest", 0));
  else
    GSenablemenuitem("quest", GSgetmenuitemname("quest", 0));
  if (City[Curr_country].pop_avail == FALSE)
    GSdisablemenuitem("quest", GSgetmenuitemname("quest", 1));
  else
    GSenablemenuitem("quest", GSgetmenuitemname("quest", 1));
  if (City[Curr_country].cost_avail == FALSE)
    GSdisablemenuitem("quest", GSgetmenuitemname("quest", 2));
  else
    GSenablemenuitem("quest", GSgetmenuitemname("quest", 2));
  sprintf(MapFile, "%s%s%s", Datapath, City[Curr_country].name, BDRSUFFIX);
  sprintf(WarpMapFile, "%s%s%s", Datapath, City[Curr_country].warp, BDRSUFFIX);
  sprintf(CityFile, "%s%s%s", Datapath, City[Curr_country].name, CITYSUFFIX);
  GSdrawnamedwindow("map", FALSE, NO_POPUP);
  GSdrawnamedwindow("cityname", FALSE, NO_POPUP);
  draw_map(MapFile, FALSE);
  return(TRUE);
}


/*----------------------------------------------------------------------
  Quit button.  Close up graphics and exit.
  GLOBALS: City, Num_cities
----------------------------------------------------------------------*/
boolean doquit(void)
{
  int i;

  for (i = 0; i < City[Curr_country].num; i++)
    /*GSelog(0, "City: %s\tlabx: %d\tlaby: %d\n",
           City[Curr_country].clist[(int)Answers[i].citynum].name,
           City[Curr_country].clist[(int)Answers[i].citynum].labx - 1,
           City[Curr_country].clist[(int)Answers[i].citynum].laby - 15);*/

  if (Move_labels)
    dump_db_file();
  GSquit(NULL);
  return(TRUE);  /* to make the compiler happy */
}


/***********************************************************************
Show user results for selected city.
GLOBALS: Answers, City, NotInStepThru, Question_item, ResultWinActive
***********************************************************************/
void show_results(int pickedcity)
{
  BigString astring;
  int dumx, dumy, dumwd, dumht;
  GSColor color;
  int storedcity;
  int y;

  storedcity = (int)Answers[pickedcity].citynum;
  ResultWinCity = pickedcity;
  if (ResultWinActive)
    GSredrawnamedwindow("Results");
  else
    GSdrawnamedwindow("Results", Use_bstore, POPUP);
  GSdisplayhelpmessage(23);
  ResultWinActive = TRUE;
  y = 5 + SMALL;
  GSgetfieldrect("Results", &dumx, &dumy, &dumwd, &dumht, &color);
  GSsetcolor(color);
  GSsetfont(SMALL);
  GSwwritemsg(4, y, City[Curr_country].clist[storedcity].name);
  y += 20;
  sprintf(astring, "Correct");
  GSwwritemsg(174-GStextwidth(astring), y, astring);
  sprintf(astring, "Your");
  GSwwritemsg(234-GStextwidth(astring), y, astring);
  y += 10;
  sprintf(astring, "Answer");
  GSwwritemsg(174-GStextwidth(astring), y, astring);
  sprintf(astring, "Answer");
  GSwwritemsg(234-GStextwidth(astring), y, astring);
  y += 5;
  GSwdrawline(174-GStextwidth(astring), y, 174, y);
  GSwdrawline(234-GStextwidth(astring), y, 234, y);
  y += 20;
  if (Question_item[TEMPERATURE] && City[Curr_country].hi_avail)
  {
    sprintf(astring, "  High Temperature: ");
    GSwwritemsg(4, y, astring);
    sprintf(astring, "%d", City[Curr_country].clist[storedcity].hi);
    GSwwritemsg(174-GStextwidth(astring), y, astring);
    if (Answers[pickedcity].hi > 0)
      sprintf(astring, "%ld", Answers[pickedcity].hi);
    else
      sprintf(astring, "%s", "none");
    GSwwritemsg(234-GStextwidth(astring), y, astring);
    y += 20;
  }
  if (Question_item[TEMPERATURE] && City[Curr_country].low_avail)
  {
    sprintf(astring, "  Low Temperature: ");
    GSwwritemsg(4, y, astring);
    sprintf(astring, "%d", City[Curr_country].clist[storedcity].low);
    GSwwritemsg(174-GStextwidth(astring), y, astring);
    if (Answers[pickedcity].low > 0)
      sprintf(astring, "%ld", Answers[pickedcity].low);
    else
      sprintf(astring, "%s", "none");
    GSwwritemsg(234-GStextwidth(astring), y, astring);
    y += 20;
  }
  if (Question_item[POPULATION] && City[Curr_country].pop_avail)
  {
    sprintf(astring, "  Population: ");
    GSwwritemsg(4, y, astring);
    sprintf(astring, "%ld", City[Curr_country].clist[storedcity].pop);
    GSwwritemsg(174-GStextwidth(astring), y, astring);
    if (Answers[pickedcity].pop > 0)
      sprintf(astring, "%ld", Answers[pickedcity].pop);
    else
      sprintf(astring, "%s", "none");
    GSwwritemsg(234-GStextwidth(astring), y, astring);
    y += 20;
  }
  if (Question_item[COST] && City[Curr_country].cost_avail)
  {
    sprintf(astring, "  Cost of living: ");
    GSwwritemsg(4, y, astring);
    sprintf(astring, "%ld", City[Curr_country].clist[storedcity].cost);
    GSwwritemsg(174-GStextwidth(astring), y, astring);
    if (Answers[pickedcity].cost > 0)
      sprintf(astring, "%ld", Answers[pickedcity].cost);
    else
      sprintf(astring, "%s", "none");
    GSwwritemsg(234-GStextwidth(astring), y, astring);
  }
}

/*---------------------------------------------------------------------
Place a label or city marker at the position of the cursor when the
user releases the mouse button.
GLOBALS: City, Got_label, LabelBeingMoved, MovedLabel,
         Xorx, Xory, Xorwd, Xorht
---------------------------------------------------------------------*/
void finish_move(void)
{
  int mfx, mfy, mfwd, mfht;
  GSColor mfcolor;

  if (Got_label)
  {
    City[Curr_country].clist[MovedLabel].labx = Xorx-1;
    City[Curr_country].clist[MovedLabel].laby = Xory+Xorht-1;
    GSgetfieldrect("Mapfield", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
    GSsetcolor(mfcolor);
    GSsetfont(SMALL);
    GSwwritemsg(Xorx+mfx-1, Xory+mfy+Xorht-1, LabelBeingMoved);
    strcpy(LabelBeingMoved, "");
    Got_label = FALSE;
  }
  else if (Got_city)
  {
    City[Curr_country].clist[MovedCity].x = Xorx+2;
    City[Curr_country].clist[MovedCity].y = Xory+Xorht-3;
    GSgetfieldrect("Mapfield", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
    GSsetcolor(mfcolor);
    draw_city_circle(City[Curr_country].clist[MovedCity].x+mfx,
                City[Curr_country].clist[MovedCity].y+mfy,
                mfcolor, NULL);
    strcpy(CityBeingMoved, "");
    Got_city = FALSE;
  }
}

/*---------------------------------------------------------------------
Move a label or city marker with the position of the cursor as the
user holds the mouse button down.
GLOBALS: City, Got_label, LabelBeingMoved, MovedLabel,
         Xorx, Xory, Xorwd, Xorht
---------------------------------------------------------------------*/
void continue_move(int x, int y)
{
  int mfx, mfy, mfwd, mfht;
  int twd, tht;
  GSColor mfcolor;

  if (Got_label)
  {
    GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSsetfont(SMALL);
    GSsetcolor(15);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
    tht = GStextheight(LabelBeingMoved);
    twd = GStextwidth(LabelBeingMoved);
    compute_xorbox(x, y, mfwd, mfht, twd, tht);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
  }
  else if (Got_city)
  {
    GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
    compute_xorbox(x, y, mfwd, mfht, 5, 5);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
  }
}

/*---------------------------------------------------------------------
  Pick up a label or city marker at the position of the cursor when
  the user presses the mouse button.
  GLOBALS: City, Got_label, LabelBeingMoved, MovedLabel,
           Xorx, Xory, Xorwd, Xorht
---------------------------------------------------------------------*/
void initiate_move(int x, int y)
{
  int i;
  int mfx, mfy, mfwd, mfht, twd, tht;
  GSColor mfcolor;

  Got_label = FALSE;
  Got_city = FALSE;
  for (i = 0; i < Move_index; i++)
  {
    if (x >= City[Curr_country].clist[i].labx &&
        x < City[Curr_country].clist[i].labx +
            GStextwidth(City[Curr_country].clist[i].name)
    &&  y <= City[Curr_country].clist[i].laby &&
        y >= City[Curr_country].clist[i].laby -
            GStextheight(City[Curr_country].clist[i].name))
    {
      MovedLabel = i;
      strcpy(LabelBeingMoved, City[Curr_country].clist[i].name);
      Got_label = TRUE;
      break;
    }
  }
  if (!Got_label)
  {
    for (i = 0; i < Move_index; i++)
    {
      if ((abs(City[Curr_country].clist[i].x - x) <= 5) &&
          (abs(City[Curr_country].clist[i].y - y) <= 5))
      {
        MovedCity = i;
        strcpy(CityBeingMoved, City[Curr_country].clist[i].name);
        Got_city = TRUE;
        break;
      }
    }
  }
  if (Got_label)
  {
    GSgetfieldrect("mapfield", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSsetcolor(15);
    GSsetfont(SMALL);
    GSwwritemsg(City[Curr_country].clist[i].labx+mfx,
                City[Curr_country].clist[i].laby+mfy, LabelBeingMoved);
    tht = GStextheight(LabelBeingMoved);
    twd = GStextwidth(LabelBeingMoved);
    tht = GStextheight(LabelBeingMoved);
    twd = GStextwidth(LabelBeingMoved);
    compute_xorbox(x, y, mfwd, mfht, twd, tht);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
  }
  else if (Got_city)
  {
    GSgetfieldrect("mapfield", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
    GSsetcolor(15);
    GSsetfont(SMALL);
    draw_city_circle(City[Curr_country].clist[i].x+mfx,
                City[Curr_country].clist[i].y+mfy, 15, NULL);
    compute_xorbox(x, y, mfwd, mfht, 5, 5);
    GSdrawxorbox(Xorx+mfx, Xory+mfy, Xorwd, Xorht);
  }
  else
  {
    GSbell();
    return;
  }
}

/*---------------------------------------------------------------------
  Drag function for country map, used in location quiz
  GLOBALS:
---------------------------------------------------------------------*/
boolean dragmap(int x, int y, int status)
{
  int cx, cy, cwd, cht;
  GSColor clabel;
  int i, pickedcity;

  if (status == MOUSE_CLICK || (status == DRAG_FINAL && !Move_labels))
  { /*--- The only time we take action ---*/
    if (Show_results)
    { /*--- Display the results for this city ---*/
      for (i = 0; i < Num_cities; i++)
	if ((abs(Citypoints[i].x - x) <= 5) && (abs(Citypoints[i].y - y) <= 5))
        {
          pickedcity = i;
          break;
        }
      if (i == Num_cities)
      {
        GSbell();
        return(TRUE);
      }
      show_results(pickedcity);
    }
    else if (Move_labels)
      GSbell();
    else
    { /* Put up the circle and label for this city where the user clicked */
      GSsetcurrentnamedwindow("Map");
      GSgetfieldrect("Usercolor", &cx, &cy, &cwd, &cht, &clabel);
      draw_city_circle(x+cx, y+cy, clabel,
                       &Citycirclepixels[Citycnt-Levelsback]);
      /* Store the user's coordinates */
      Citypoints[Citycnt-Levelsback].ux = x;
      Citypoints[Citycnt-Levelsback].uy = y;
      GSsetfont(LARGE);
      /* Now, return control to askcity for the next question */
      askcity();
    }
  }
  else if ((status == DRAG_PROCESS) && (Move_labels))
  {
    continue_move(x, y);
  }
  else if ((status == DRAG_INIT) && (Move_labels))
  {
    initiate_move(x, y);
  }
  else if ((status == DRAG_FINAL) && (Move_labels))
  {
    finish_move();
  }
  return(TRUE);
}


/***********************************************************************
  Generic slider drag function
  GLOBALS: Currslide
           Slideval
***********************************************************************/
boolean dragslide(int relx, int rely, int status)
{
  BigString string;
  long value;
  double fract;
  int offset;
  static int offx;
  int offy, offwd, offht;
  GSColor offc;

  offset = rely;  /* to make the compiler happy */
  if (status == DRAG_INIT)
  {
    GSgetfieldrect("Dragbox", &offx, &offy, &offwd, &offht, &offc);
    if (Currslide.writeval)
      GSdisplayhelpmessage(18);
    else
      GSdisplayhelpmessage(19);
  }
  else if (status == DRAG_PROCESS)
  {
    GSwclearrect(Currslide.x, Currslide.y, Currslide.wd+2, Currslide.ht+1);
    GSwclearrect(Currslide.x, Currslide.y+SLVALOFFY, Currslide.wd, SMALL+2);
    fract = drawslider(Currslide.x, Currslide.y, Currslide.wd, Currslide.ht,
                       Currslide.barcolor, Currslide.tickcolor, relx+offx);
    if (Currslide.writeval)
    {
      value = (long)(fract * Currslide.range + 0.5) + Currslide.min;
      sprintf(string, "%3ld%s", value, Units);
      offset = (Currslide.wd - GStextwidth(string))/2;
      *Slideval = value;
      GSsetfont(SMALL);
      GSwwritemsg(Currslide.x+offset, Currslide.y+SLVALOFFY+SMALL, string);
    }
    else
      *Slideval = (long)(fract * 100);
  }
  else if (status == DRAG_FINAL)
    GSdisplayhelpmessage(21);
  return(TRUE);
}


/***********************************************************************
Returns user to previous city in quiz.
GLOBALS: Citycnt
         Levelsback
         Quest_status
         Questesc
***********************************************************************/
boolean goback(void)
{
  if ((Levelsback < Citycnt) || Questesc)
  {
    QuestionWinActive = FALSE;
    Quest_status = RELOCATION;
    askcity();
  }
  else
    GSbell();
  return(TRUE);
}


/***********************************************************************
Takes user to next city in step thru at end of quiz.
GLOBALS: Citycnt
         Levelsback
         Quest_status
         Questesc
***********************************************************************/
boolean go_on(void)
{
  step_thru();
  return (TRUE);
}


/***********************************************************************
Show message giving instructions for popup "Help" menu
GLOBALS:
***********************************************************************/
boolean help(void)
{
  GSdisplayhelpmessage(9); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/***********************************************************************
Finished looking at a help file
GLOBALS: NotInStepThru
         Use_bstore
         Default_country
         Curr_country
         Savemsg
***********************************************************************/
boolean helpdone(void)
{
  if (HelpWinActive)
    GSremovenamedwindow("help");
  HelpWinActive = FALSE;
  if (!Use_bstore)
    redraw_all();
  /* start quiz on default country if one was specified in config file */
  if (Default_country >= 0)
  {
    Curr_country = Default_country;
    Default_country = -1; /* reset, so this only happens once */
    return(countryitem(GSgetmenuitemname("Country", Curr_country)));
  }
  GSdisplayhelpmessage(Savemsg);
  return(TRUE);
}


/***********************************************************************
"Next" button on a help screen
GLOBALS: Helpfile
         Use_bstore
***********************************************************************/
boolean helpnext(void)
{
  return(GSdohelpscreens(Helpfile, NextPage, Use_bstore, CURR_MSG));
}


/***********************************************************************
"Prev" button on a help screen
GLOBALS: Helpfile
         Use_bstore
***********************************************************************/
boolean helpprev(void)
{
  return(GSdohelpscreens(Helpfile, PrevPage, Use_bstore, CURR_MSG));
}


/***********************************************************************
Show message giving instructions for popup "Info" menu
GLOBALS:
***********************************************************************/
boolean info(void)
{
  GSdisplayhelpmessage(12); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/***********************************************************************
Display help message when "Number" button is pressed
GLOBALS:
***********************************************************************/
boolean number(void)
{
  GSdisplayhelpmessage(5); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/***********************************************************************
Handle an item in the "Number" (of cities) menu.
Remove check from old value, and then set the new number of cities.
GLOBALS: Num_cities
***********************************************************************/
boolean numberitem(char *itemname)
{
	int currcheck;
	
	int itemnum = GSgetmenuitemnum("Number", itemname);
	
	if (Num_cities == 68)
	currcheck = 4;
	else
	currcheck = Num_cities/5 -1;
	
	GStogglecheck("Number", GSgetmenuitemname("Number", currcheck));
	Num_cities = (itemnum+1)*5;
	return(TRUE);
}


/***********************************************************************
Show message giving instructions for popup "Questions" menu
GLOBALS:
***********************************************************************/
boolean quest(void)
{
  GSdisplayhelpmessage(14); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/***********************************************************************
"Done" button in a question box
GLOBALS:
***********************************************************************/
boolean questdone(void)
{
  int x, y;
  GSColor color;

/*  GSremovenamedwindow("Question"); / It should be using backing store /
  QuestionWinActive = FALSE;
  if (!Use_bstore)
  {
    redraw_all();
    GSsetcurrentnamedwindow("map");
    GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
    draw_city_circle(Citypoints[Citycnt-Levelsback].ux+x,
                     Citypoints[Citycnt-Levelsback].uy+y, color, NULL);
  }*/

  GSsetcurrentnamedwindow("Map");
  askcity();
  return(TRUE);
}


/***********************************************************************
Escape question window and go back one city
GLOBALS: Citycnt
         Levelsback
         Questesc
***********************************************************************/
boolean questesc(void)
{
  GSremovenamedwindow("Question"); /* It should be using backing store */
  QuestionWinActive = FALSE;
  if (!Use_bstore)
    redraw_all();
  GSsetcurrentnamedwindow("Map");
  Questesc = TRUE;
/*  if (Citycnt >= Levelsback)
    GSenablebutton("cityname", "GoBack", DRAW);*/
  return(goback());
}


/***********************************************************************
Show message giving instructions for popup "Quit" menu
GLOBALS:
***********************************************************************/
boolean quit(void)
{
  GSdisplayhelpmessage(11); /* don't Savemsg for popup menu instructions */
  return(TRUE);
}


/***********************************************************************
Done looking at results window
GLOBALS: NotInStepThru
***********************************************************************/
boolean resdone(void)
{
  GSremovenamedwindow("Results");       /* It should be using backing store */
  GSdisplayhelpmessage(Savemsg);
  ResultWinActive = FALSE;
  if (!Use_bstore) redraw_all();
/*  GSenablebutton("root", "file", DRAW);
  GSenablebutton("root", "info", DRAW);
  GSenablebutton("root", "help", DRAW);
  GSenablebutton("cityname", "start", DRAW);*/
  if (!NotInStepThru)
  {
/*    GSenablebutton("message", "go_on", DRAW);*/
  }
  else
  {
/*    GSenablebutton("root", "country", DRAW);
    GSenablebutton("root", "number", DRAW);
    GSenablebutton("root", "quest", DRAW);*/
    GSsetcurrentnamedwindow("Map");
  }
  return(TRUE);
}


/***********************************************************************
Done looking at summary table window
GLOBALS: SummaryTableWinActive
***********************************************************************/
boolean summtabdone(void)
{
  GSremovenamedwindow("SummaryTable");  /* It should be using backing store */
  SummaryTableWinActive = FALSE;
  if (!Use_bstore) redraw_all();
  GSdisablebutton("SummaryTable", "summtabdone", NO_DRAW);
  show_final_map();
  if (Show_results)    
    GSdisplayhelpmessage(Savemsg = 8);
  else
    GSdisplayhelpmessage(Savemsg = 13);
  GSenablebutton("root", "number", DRAW);
  GSenablebutton("root", "quest", DRAW);
  GSenablebutton("root", "country", DRAW);
  return(TRUE);
}


/***********************************************************************
Start a round of cities
GLOBALS: Move_labels
         Questesc
         StepCount
         Levelsback
         NotInStepThru
         Show_results
         Answers
         Curr_city
         MapFile
         City
         ReadDebugInput
         Citycnt
***********************************************************************/
boolean start(void)
{
  int i;
  int mx, my, mwd, mht;
  GSColor mcolor;

  FinalMapActive = FALSE;
  QuestionWinActive = FALSE;
  HelpWinActive = FALSE;
  SummaryTableWinActive = FALSE;
  ResultWinActive = FALSE;
  for (i = 0; i < 4; i++)
    GSenablemenuitem("number", GSgetmenuitemname("number", i));
  GSdisablebutton("cityname", "GoBack", DRAW);
  GSinvisiblebutton("Message", "go_on", DRAW);
  GSinvisiblebutton("Message", "next", DRAW);
  GSenablebutton("root", "number", DRAW);
  GSenablebutton("root", "quest", DRAW);
  GSenablebutton("root", "country", DRAW);
  Move_labels = FALSE;
  Questesc = FALSE;
  StepCount = 0;
  Levelsback = 0;
  NotInStepThru = TRUE;
  Show_results = FALSE;
  QuizInProgress = TRUE;
  for (i = 0 ; i < MAXASK ; i++)
  {
    Answers[i].hi = Answers[i].low = Answers[i].pop = Answers[i].cost =
    Answers[i].pref = -1;
  }
  if (strcmp(MapFile, "") == 0)
  {
    GSalert(1, Use_bstore, "nomapfile1", "nomapfile2", "");
    AlertNumber = 3;
  }
  else
  {
    if (Curr_city != 0)
    { /* Clear the screen and redraw the map */
      if (Default_country >= 0)
      {
        Curr_country = Default_country;
        Default_country = -1; /* reset, so this only happens once */
        sprintf(MapFile, "%s%s%s", Datapath, City[Curr_country].name,
                BDRSUFFIX);
        sprintf(WarpMapFile, "%s%s%s", Datapath, City[Curr_country].warp,
                BDRSUFFIX);
        sprintf(CityFile, "%s%s%s", Datapath, City[Curr_country].name,
                CITYSUFFIX);
      }
      GSsetcurrentnamedwindow("Map");
      GSgetfieldrect("MapField", &mx, &my, &mwd, &mht, &mcolor);
      GSwclearrect(mx, my, mwd, mht);
      draw_map(MapFile, FALSE);
    }
    /* Clear the marks for the last quiz from the city record array */
    for(i=0; i<City[Curr_country].num; i++)
      City[Curr_country].clist[i].mark = FALSE;
    for(i=0; i<MAXASK; i++)
      Answers[i].citynum = Answers[i].hi = Answers[i].low =
      Answers[i].pop = Answers[i].cost = Answers[i].pref = -2;
    GSsetactivenamedwindow("Map");
    GSenabledragarea("Map", "Dragmap");
  if(ReadDebugInput || Moving)
  {
    finish_quiz();
    return(TRUE);
  }
    Citycnt = -1;
    askcity();
  }
  return(TRUE);
}

/*------------------------------------------------------------------------------
  login_return_func: passed to GSdologinscreens to be executed when user hits
                     RETURN
------------------------------------------------------------------------------*/
boolean login_return_func (
  void
)
{
  LoginInProcess = FALSE;
  GSdrawnamedwindow("root", FALSE, NO_POPUP);
  GSdisplayversion(VersionFile);
  GSdrawnamedwindow("message", FALSE, NO_POPUP);
  HelpWinActive = TRUE;
  return(GSdohelpscreens(Helpfile, FirstPage, Use_bstore, 1));
}

/*------------------------------------------------------------------------------
Display the version number and then the introductory help screen
GLOBALS: Infopath
         Helpfile
         Use_bstore
         VersionFile
------------------------------------------------------------------------------*/
boolean startup(void)
{
  int i;

  GSdeactivatefunction ("startup");
  for (i = Num_countries; i < 4; i++)
  {
    GSdisablemenuitem ("country", GSgetmenuitemname("country", i));
  }
  sprintf (Helpfile, "%s%s", Infopath, INTRO_HLP);
  if ( GetName == TRUE )
  {
    LoginInProcess = TRUE;
    GSdologinscreen ("Please type your name (First Last) and then hit Return:",
                     NO_BSTORE,
                     login_return_func,
                     GSconvertcolor("loginfg"),
                     (char *)UserName);
    GSdisplayversion ( VersionFile ) ;
    return ( TRUE ) ;
  }
  else if ( Moving == TRUE )
  {
    /*--- if moving cities, we presume that initial help not needed ---*/
    redraw_all();
    return ( helpdone ( ) ) ;
  }
  else
  {
  	redraw_all();
    HelpWinActive = TRUE;
    return ( GSdohelpscreens ( Helpfile , FirstPage , Use_bstore , 1 ) ) ;
  }
}


/***********************************************************************
Toggle a question menu item
GLOBALS: Question_item
***********************************************************************/
boolean toggleq(char *itemname)
{
  int itemno = GSgetmenuitemnum("Quest", itemname);
  
  Question_item[itemno] = !Question_item[itemno];
  if (Question_item[itemno]) {
    GSdisplayhelpmessage(7); /* clear window in case message 15 is currently
                                displayed; if you don't, it looks like help
                                message has not responded to user input */
    GSdisplayhelpmessage(15);
  }
  else {
    GSdisplayhelpmessage(7); /* clear window in case message 15 is currently
                                displayed; if you don't, it looks like help
                                message has not responded to user input */
    GSdisplayhelpmessage(16);
  }
  return(TRUE);
}


/***********************************************************************
Initiate main help window
GLOBALS: Helpfile
         Infopath
         NotInStepThru
         Use_bstore
***********************************************************************/
boolean tutorcmd(void)
{
  sprintf(Helpfile, "%s%s", Infopath, INTRO_HLP);
  HelpWinActive = TRUE;
  return(GSdohelpscreens(Helpfile, FirstPage, Use_bstore, 10));
}


/***********************************************************************
Initiate "Version" help screen
GLOBALS: Helpfile
         Infopath
         NotInStepThru
         Use_bstore
***********************************************************************/
boolean versioncmd(void)
{
  sprintf(Helpfile, "%s%s", Infopath, VERSION_INFO);
  HelpWinActive = TRUE;
  return(GSdohelpscreens(Helpfile, FirstPage, Use_bstore, 10));
}

/*-------------------------------------------------------------------------------
  Print screen to file
-------------------------------------------------------------------------------*/
boolean dumpscreen(void)
{
  BigString msgstring;
  int x, y;
  GSColor color;

  GSremovenamedwindow("dump");
  GSdisplayhelpmessage(Savemsg);
  DumpWinActive = FALSE;
  if (!Use_bstore)
    redraw_all();
  strcpy (DumpFileName, "mmap");
  if (GraphFileFmt != NO_FILE)
    DumpSuccess = GSdumpscreen(DumpFileName, GraphFileFmt);
  else return FALSE;

  GSdisplayhelpmessage(1);
  DumpResultsActive = TRUE;
  GSdrawnamedwindow("alert", Use_bstore, POPUP);
  GSsetcolor(GSconvertcolor("black"));
  GSsetfont(LARGE);

  if (DumpSuccess)
  {
    sprintf(msgstring, "The screen was saved to file:");
    GSwwritemsg(6, 30, msgstring);
    sprintf(msgstring, "   %s", DumpFileName);
    GSwwritemsg(6, 45, msgstring);
    return TRUE;
  }
  else
  {
    sprintf(msgstring, "An error occurred saving the screen.");
    GSwwritemsg(6, 15, msgstring);
    sprintf(msgstring, "No file was saved.");
    GSwwritemsg(6, 30, msgstring);
    sprintf(msgstring, "Check disk space and file permissions.");
    GSwwritemsg(6, 45, msgstring);
    GSbell();
    return FALSE;
  }
}

void highlightfield (
  BigString  window,
  BigString  field,
  GSColor    bgcolor,
  GSColor    topleft,
  GSColor    bottomright
)
{
  int fx, fy, fwd, fht;

  GSgetfieldrect(field, &fx, &fy, &fwd, &fht, NULL);
  GSsetcurrentnamedwindow(window);
  GSsetcolor(bgcolor);
  GSwfillrect(fx, fy, fwd, fht);
  GSsetcolor(topleft);
  GSwdrawline(fx-1, fy-1, fx+fwd, fy-1);
  GSwdrawline(fx-2, fy-2, fx+fwd+1, fy-2);
  GSwdrawline(fx-1, fy-1, fx-1, fy+fht);
  GSwdrawline(fx-2, fy-2, fx-2, fy+fht+1);
  GSsetcolor(bottomright);
  GSwdrawline(fx, fy+fht, fx+fwd, fy+fht);
  GSwdrawline(fx-1, fy+fht+1, fx+fwd+1, fy+fht+1);
  GSwdrawline(fx+fwd, fy, fx+fwd, fy+fht);
  GSwdrawline(fx+fwd+1, fy-1, fx+fwd+1, fy+fht+1);
}

/*-------------------------------------------------------------------------------
  Displays the "dump" window, warning the user that this will take a long
  time. The user is asked to confirm his or her wish to dump the screen to file
-------------------------------------------------------------------------------*/
boolean confirmdump (
  void
)
{
  int x, y;
  GSColor color;

  GSdisplayhelpmessage(28);
  DumpWinActive = TRUE;
  GSdrawnamedwindow("dump", Use_bstore, POPUP);
  GSgetfieldrect("filetype", &x, &y, NULL, NULL, &color);
  highlightfield("dump", "filetype", GSconvertcolor("lgray"),
                 GSconvertcolor("black"), GSconvertcolor("white"));
  GSsetcolor(color);
  GSsetfont(LARGE);
  switch (GraphFileFmt) {
    case BMP_16:
      GSwwritenamedmsg(x+5, y+SMALL+2, "bmp16msg");
      break;
    case GIF_16:
      GSwwritenamedmsg(x+5, y+SMALL+2, "gif16msg");
      break;
    case PS_GRAY:
      GSwwritenamedmsg(x+5, y+SMALL+2, "psgraymsg");
      break;
    case PS_COLOR:
      GSwwritenamedmsg(x+5, y+SMALL+2, "pscolormsg");
      break;
    case NO_FILE: default:
      GSwwritenamedmsg(x+5, y+SMALL+2, "nofilemsg");
      break;
  }
  return TRUE;
}

/*-------------------------------------------------------------------------------
  Attached to cancel screen dump button in dump window.
-------------------------------------------------------------------------------*/
boolean canceldump (
  void
)
{
  GSremovenamedwindow("dump");
  DumpWinActive = FALSE;
  if (!Use_bstore)
    redraw_all();
  GSdisplayhelpmessage (Savemsg);
  return TRUE;
}

/*-------------------------------------------------------------------------------
  set_format: set a graphics file format for screen dump in response to user
              selection from GFformat popup menu
-------------------------------------------------------------------------------*/
boolean set_format (
	char *itemname
)
{
  int x, y;
  GSColor color;
  int item_n = GSgetmenuitemnum("GFformat", itemname);

  GSgetfieldrect("filetype", &x, &y, NULL, NULL, &color);
  highlightfield("dump", "filetype", GSconvertcolor("lgray"),
                 GSconvertcolor("black"), GSconvertcolor("white"));
  GSsetcolor(color);
  GSsetfont(LARGE);
  switch (item_n) {
    case 0:
      GraphFileFmt = BMP_16;
      GSwwritenamedmsg(x+5, y+SMALL+2, "bmp16msg");
      break;
    case 1:
      GraphFileFmt = PS_COLOR;
      GSwwritenamedmsg(x+5, y+SMALL+2, "pscolormsg");
      break;
    case 2:
      GraphFileFmt = PS_GRAY;
      GSwwritenamedmsg(x+5, y+SMALL+2, "psgraymsg");
      break;
    case 3:
      GraphFileFmt = GIF_16;
      GSwwritenamedmsg(x+5, y+SMALL+2, "gif16msg");
      break;
  }
  return (TRUE);
}

/*-----------------------------------------------------------------------------
  gfformat - associated with popup menu which gives user a choice of graphic
             file formats to dump to.
-----------------------------------------------------------------------------*/
boolean gfformat (
  void
)
{
  GSdisplayhelpmessage(30);
  return (TRUE);
}

/*------------------------------------------------------------------------------
  file  - associated with 'File' popup menu.
------------------------------------------------------------------------------*/
boolean file (
  void
)
{
  GSdisplayhelpmessage(29);
  return (TRUE);
}

/* Array associating string names with function pointers for buttons */
IntrFunction FuncNames[] =
{
  { "aboutq",            aboutq      },
  { "alertok",           alertok     },
  { "canceldump",        canceldump  },
  { "cancelquit",        cancelquit  },
  { "confirmdump",       confirmdump },
  { "confirmquit",       confirmquit },
  { "dumpscreen",        dumpscreen  },
  { "country",           countryhelp },
  { "countryitem",       countryitem },
  { "doquit",            doquit      },
  { "dragmap",           dragmap     },
  { "dragslide",         dragslide   },
  { "file",              file        },
  { "gfformat",          gfformat    },
  { "goback",            goback      },
  { "go_on",             go_on       },
  { "help",              help        },
  { "helpdone",          helpdone    },
  { "helpnext",          helpnext    },
  { "helpprev",          helpprev    },
  { "info",              info        },
  { "login_return_func", login_return_func },
  { "nextcity",          do_moving   },
  { "number",            number      },
  { "numberitem",        numberitem  },
  { "quest",             quest       },
  { "questdone",         questdone   },
  { "questesc",          questesc    },
  { "quit",              quit        },
  { "redraw_all",        redraw_all  },
  { "resdone",           resdone     },
  { "set_format",        set_format  },
  { "summtabdone",       summtabdone },
  { "start",             start       },
  { "startup",           startup     },
  { "toggleq",           toggleq     },
  { "tutorcmd",          tutorcmd    },
  { "versioncmd",        versioncmd  },
};

int NumFuncs = GSnumifuncs(FuncNames);
