/***********************************************************************
Mental Maps module: main.c, configuration and initialization.
***********************************************************************/
#include "geosim.h"

#include <stdlib.h>
#include <stdio.h>
#include <time.h>   /* For randomize */
#ifdef MSDOS
#include <malloc.h>
#endif
#include <string.h>
#ifndef MAXLONG
#include <values.h> /* MAXLONG is the only thing needed from this header.
			On the Mac, MAXLONG is declared in the gssys.h header. JBR*/
#endif
#include "mmincl.h"

FileName Config_name = "config.gcf"; /* name of configuration file */
				     /* must be in directory with mmap.exe */


/***********************************************************************
Local reader for a city list file.
GLOBALS:
***********************************************************************/
boolean city_reader(
  BigString buffer,  /* string to process */
  ReaderStatusType status,        /* what to do */
  void *data
)
{
  static int numcities = 0;      /* total number of cities in .cty file */
  BigString temp1;
  BigString temp2;
  int numkwds;                   /* number of keywords in format statement */
  FormatList format;             /* list to hold keywords in format statement */
  static double mapscale;
  static double fudgex = 1.0;
  static double fudgey = 1.0;
  int i;

  data = data; /*--- to quiet compiler ---*/
  if (status == READ_INIT)
  {
    mapscale = MAPSCALE;
    fudgex = fudgey = 0.0;
    return(TRUE);
  }
  if (status == READ_FINAL)
  {
    numcities = 0; /* Reset to check number of cities in next file */
    return(TRUE);
  }

  if (GSstricmp(buffer, "mapscale=") == 0)
  {
    mapscale = GSrdreal();
  }
  if (GSstricmp(buffer, "mapoffx=") == 0)
  {
    fudgex = GSrdreal();
    City[Curr_country].fudgex = fudgex;
  }
  if (GSstricmp(buffer, "mapoffy=") == 0)
  {
    fudgey = GSrdreal();
    City[Curr_country].fudgey = fudgey;
  }
  else if (GSstricmp(buffer, "numcities=") == 0)
  {
    if (numcities != 0)
    {
      GSelog(0, "Numcities not 0 in city data file\n");
      return(FALSE);
    }
    City[Curr_country].num = numcities = (int)GSrdinteger();
    if ((City[Curr_country].clist =
           (Cityrec *)malloc((long)City[Curr_country].num *
                             sizeof (Cityrec))) == NULL)
      GSelog(-1, "Out of memory for city records\n");
  }
  else if (strcmp(buffer, "format=") == 0)
  { /* Process format statement */
    if (numcities == 0) /* Should have seen this by now */
     GSelog(-1, "Error in city file - missing numcities\n");
    if (!GSrdformat(&numkwds, format))
      return(FALSE);
    if ((strcmp(format[0], "x") != 0) || (strcmp(format[1], "y") != 0))
    {
      GSelog(0, "Bad format statement\n");
      return(FALSE);
    }
    i = 2;
    if (strcmp(format[i], "pop") != 0 || i >= numkwds)
      City[Curr_country].pop_avail = FALSE;
    else
    {
      City[Curr_country].pop_avail = TRUE;
      i++;
    }
    if (strcmp(format[i], "low") != 0 || i >= numkwds)
      City[Curr_country].low_avail = FALSE;
    else
    {
      City[Curr_country].low_avail = TRUE;
      i++;
    }
    if (strcmp(format[i], "hi") != 0 || i >= numkwds)
      City[Curr_country].hi_avail = FALSE;
    else
    {
      City[Curr_country].hi_avail = TRUE;
      i++;
    }
    if (strcmp(format[i], "cost") != 0 || i >= numkwds)
      City[Curr_country].cost_avail = FALSE;
    else
    {
      City[Curr_country].cost_avail = TRUE;
      i++;
    }
    if (strcmp(format[i], "name") != 0 || i >= numkwds)
    {
      GSelog(0, "Bad format statement\n");
      return(FALSE);
    }
    else
      i++;
    if (strcmp(format[i], "st") != 0 || i >= numkwds)
      City[Curr_country].st_avail = FALSE;
    else
    {
      City[Curr_country].st_avail = TRUE;
      i++;
    }
    if (strcmp(format[i], "labx") != 0 || i >= numkwds)
    {
      if (!Moving)
      {
        GSelog(0, "Bad format statement\n");
        return(FALSE);
      }
      City[Curr_country].labx_avail = FALSE;
    }
    else
    {
      i++;
      City[Curr_country].labx_avail = TRUE;
    }
    if (strcmp(format[i], "laby") != 0 || i >= numkwds)
    {
      if (!Moving)
      {
        GSelog(0, "Bad format statement\n");
        return(FALSE);
      }
      City[Curr_country].laby_avail = FALSE;
    }
    else
    {
      i++;
      City[Curr_country].laby_avail = TRUE;
    }
    for (Curr_city=0; Curr_city < City[Curr_country].num; Curr_city++)
    {
      City[Curr_country].clist[Curr_city].x = (int)(GSrdreal() * mapscale +
                                                    fudgex);
      City[Curr_country].clist[Curr_city].y = (int)(GSrdreal() * mapscale +
                                                    fudgey);
      if (City[Curr_country].pop_avail)
        City[Curr_country].clist[Curr_city].pop = GSrdinteger();
      if (City[Curr_country].low_avail)
        City[Curr_country].clist[Curr_city].low = (int)GSrdinteger();
      if (City[Curr_country].hi_avail)
        City[Curr_country].clist[Curr_city].hi = (int)GSrdinteger();
      if (City[Curr_country].cost_avail)
        City[Curr_country].clist[Curr_city].cost = (int)(GSrdreal() + 0.5);
      if (City[Curr_country].st_avail)
      { /* Get both city and state name */
        if(!GSrdphrase(temp1, BIGSTRINGLEN))
          return(FALSE);
        if(!GSrdphrase(temp2, BIGSTRINGLEN))
          return(FALSE);
        sprintf(City[Curr_country].clist[Curr_city].name, "%s, %s", temp1,
                temp2);
      }
      else
        if(!GSrdphrase(City[Curr_country].clist[Curr_city].name, BIGSTRINGLEN))
          return(FALSE);
      if (City[Curr_country].labx_avail)
        City[Curr_country].clist[Curr_city].labx =
          (int)(GSrdreal() * mapscale + fudgex);
      else
        City[Curr_country].clist[Curr_city].labx = 0;
      if (City[Curr_country].laby_avail)
        City[Curr_country].clist[Curr_city].laby =
         (int)(GSrdreal() * mapscale + fudgey);
      else
        City[Curr_country].clist[Curr_city].laby = 0;
    }
    if(!GSrdword(temp1, BIGSTRINGLEN))
      return(FALSE);
    if (strcmp(temp1, "endformat") != 0)
      return(FALSE);
  }
  return(TRUE);
}


/***********************************************************************
Local reader for the configuration file.
GLOBALS: Config_name, Home, InterfaceFile, Datapath, Infopath
***********************************************************************/
boolean config_reader(
  BigString buffer,  /* string to process */
  ReaderStatusType status,        /* what to do */
  void *data
)
{
  BigString tmpstr;
  BigString filename;
  FILE *city_fp;  /* A country's city list file */
  int i;

  data = data; /*--- to quiet compiler ---*/
  if (status == READ_INIT)
    return(TRUE);  /* nothing to do */
  if (status == READ_FINAL)
    return(TRUE);  /* nothing to do */

  if (GSstricmp(buffer, "datapath=") == 0)
  {
    if(!GSrdword(tmpstr, BIGSTRINGLEN))
      return(FALSE);
    sprintf(Datapath, "%s%s", Home, tmpstr);
  }
  else if (GSstricmp(buffer, "graphfileformat=") == 0)
  {
    if(!GSrdword(tmpstr, BIGSTRINGLEN))
    {
      return(FALSE);
    }
    if (GSstricmp(tmpstr, "BMP16") == 0)
    {
      GraphFileFmt = BMP_16;
    }
    else if (GSstricmp(tmpstr, "PSGREY") == 0)
    {
      GraphFileFmt = PS_GRAY;
    }
    else if (GSstricmp(tmpstr, "PSCOLOR") == 0)
    {
      GraphFileFmt = PS_COLOR;
    }
    else if (GSstricmp(tmpstr, "GIF16") == 0)
    {
      GraphFileFmt = GIF_16;
    }
    else
    {
      GraphFileFmt = NO_FILE;
    }
  }
  else if (GSstricmp(buffer, "infopath=") == 0)
  {
    if(!GSrdword(tmpstr, BIGSTRINGLEN))
      return(FALSE);
    sprintf(Infopath, "%s%s", Home, tmpstr);
  }
  else if (GSstricmp(buffer, "interfacefile=") == 0)
  {
    if(!GSrdword(tmpstr, BIGSTRINGLEN))
      return(FALSE);
    sprintf(InterfaceFile, "%s%s", Home, tmpstr);
  }
  else if (GSstricmp(buffer, "getname=") == 0)
    GetName = GSrdboolean();
  else if (GSstricmp(buffer, "bstore=") == 0)
    Use_bstore = GSrdboolean();
  else if (GSstricmp(buffer, "outputlogmode=") == 0)
    GSrdstringeol(Outputlogmode, BIGSTRINGLEN);
  else if (GSstricmp(buffer, "numcities=") == 0)
    Num_cities = (int)GSrdinteger();
  else if (GSstricmp(buffer, "preference=") == 0)
    Question_item[PREF] = GSrdboolean();
  else if (GSstricmp(buffer, "temperature=") == 0)
    Question_item[TEMPERATURE] = GSrdboolean();
  else if (GSstricmp(buffer, "population=") == 0)
    Question_item[POPULATION] = GSrdboolean();
  else if (GSstricmp(buffer, "costofliving=") == 0)
    Question_item[COST] = GSrdboolean();
  else if (GSstricmp(buffer, "stepthru=") == 0)
    StepThru = GSrdboolean();
  else if (GSstricmp(buffer, "readdebug=") == 0)
    ReadDebugInput = GSrdboolean();
  else if (GSstricmp(buffer, "moving=") == 0)
    Moving = GSrdboolean();
  else if (GSstricmp(buffer, "defaultcountry=") == 0)
  {
    if(!GSrdword(tmpstr, BIGSTRINGLEN))
      return(FALSE);
    if(GSstricmp(tmpstr, "usa") == 0)
      Default_country = 0;
    else if(GSstricmp(tmpstr, "usast") == 0)
      Default_country = 1;
    else if(GSstricmp(tmpstr, "india") == 0)
      Default_country = 2;
    else if(GSstricmp(tmpstr, "mideast") == 0)
      Default_country = 3; /* -dh was and s/b 3 */
  }
  else if (GSstricmp(buffer, "numcountries=") == 0)
  {
    Curr_country = -1;
    Num_countries = (int)GSrdinteger();
    if ((City = (Countryrec *)malloc((long)Num_countries *
                                      sizeof (Countryrec))) == NULL)
      GSelog(-1, "Out of memory for country records\n");
  }
  else if (GSstricmp(buffer, "countryfile=") == 0)
  {
    Curr_country++;
    if (Curr_country == Num_countries)
      GSelog(-1, "Error with a country file - too many countries?\n");
    if(!GSrdword(City[Curr_country].name, FILEPREFIXLEN))
      GSelog(-1, "Error with reading config file word |%s|\n",
             City[Curr_country].name);
    strcpy(City[Curr_country].warp, City[Curr_country].name);
    sprintf(filename, "%s%s%s", Datapath, City[Curr_country].name, CITYSUFFIX);
    if ((city_fp = GSopen(filename, "r")) == NULL)
      GSelog(-1, "Cannot open file |%s|\n", filename);
    if (!GSgenreader(city_fp, city_reader, "MM_cdata", NULL))
      GSelog(-1, "Error while reading city file |%s|\n", filename);
    fclose(city_fp);
    /* Calculate city min's and max's */
    City[Curr_country].minhi = MAXLONG;
    City[Curr_country].maxhi = -1;
    for(i=0; i< City[Curr_country].num; i++)
    {
      if (City[Curr_country].clist[i].hi > City[Curr_country].maxhi)
        City[Curr_country].maxhi = City[Curr_country].clist[i].hi;
      if (City[Curr_country].clist[i].hi < City[Curr_country].minhi)
        City[Curr_country].minhi = City[Curr_country].clist[i].hi;
    }
    City[Curr_country].minlow = MAXLONG;
    City[Curr_country].maxlow = -1;
    for(i=0; i< City[Curr_country].num; i++)
    {
      if (City[Curr_country].clist[i].low > City[Curr_country].maxlow)
        City[Curr_country].maxlow = City[Curr_country].clist[i].low;
      if (City[Curr_country].clist[i].low < City[Curr_country].minlow)
        City[Curr_country].minlow = City[Curr_country].clist[i].low;
    }
    City[Curr_country].minpop = MAXLONG;
    City[Curr_country].maxpop = -1;
    for(i=0; i< City[Curr_country].num; i++)
    {
      if (City[Curr_country].clist[i].pop > City[Curr_country].maxpop)
        City[Curr_country].maxpop = City[Curr_country].clist[i].pop;
      if (City[Curr_country].clist[i].pop < City[Curr_country].minpop)
        City[Curr_country].minpop = City[Curr_country].clist[i].pop;
    }
    City[Curr_country].mincost = MAXLONG;
    City[Curr_country].maxcost = -1;
    for(i=0; i< City[Curr_country].num; i++)
    {
      if (City[Curr_country].clist[i].cost > City[Curr_country].maxcost)
        City[Curr_country].maxcost = City[Curr_country].clist[i].cost;
      if (City[Curr_country].clist[i].cost < City[Curr_country].mincost)
        City[Curr_country].mincost = City[Curr_country].clist[i].cost;
    }
  }
  else if (GSstricmp(buffer, "warpfile=") == 0)
  {
    if(!GSrdword(City[Curr_country].warp, FILEPREFIXLEN))
      GSelog(-1, "Error with reading config file word |%s|\n",
             City[Curr_country].warp);
  }
  return(TRUE);
}


/***********************************************************************
Process the configuration file -- initiate reader.
STATIC GLOBALS: Config_name, Home.
***********************************************************************/
void configure(void)
{
  FILE *config_fp;
  BigString filename;

  sprintf(filename, "%s%s", Home, Config_name);
  if ((config_fp = GSopen(filename, "r")) == NULL)
    return; /* Unable to open the configure file - use defaults. */
  if (!GSgenreader(config_fp, config_reader, "MM_config", NULL))
    GSelog(-1, "Error while reading configuration file |%s|\n", filename);
  fclose(config_fp);
}


/**********************************************************************
Print out the usage message and exit.

GLOBALS USED: None.
**********************************************************************/
void printusage(void)
{
  printf("Mental Maps.\n");
  printf("\tUsage: mmap [-d] [-i] [-m] [-l] [-sy] [-sn] \n");
  printf("The -d option turns on debug messages\n");
  printf("The -i option reads from input.txt debug file\n");
  printf("The -m option allows movement of city labels\n");
  printf("The -l option turns on login procedure\n");
  printf("The -sy option turns on stepthru review or cities\n");
  printf("The -sn option turns off stepthru review or cities\n");
  GSdelay(ERRORDELAY);
  exit(0);
}

void process_args (
  int    argc ,
  char **argv
)
{
  int i ;
  
  for ( i = 1 ; i < argc ; i++ )
  {
    if (GSstricmp (argv[i], "-d") == 0)
      GSDebug = TRUE;
    else if (GSstricmp (argv[i], "-i") == 0)
      ReadDebugInput = TRUE;
    else if (GSstricmp (argv[i], "-m") == 0)
      Moving = TRUE;
    else if (GSstricmp (argv[i], "-l") == 0)
      GetName = TRUE;
    else if (GSstricmp (argv[i], "-sn") == 0)
      StepThru = FALSE;
    else if (GSstricmp (argv[i], "-sy") == 0)
      StepThru = TRUE;
    else
      printusage();
  }
}

/***********************************************************************
Main routine: Do initializations and start interface interpreter.
GLOBALS: None.
***********************************************************************/
#if 0
int main(int argc, char **argv)
/* argc, argv dependency removed - JBR */
#else
int GS_MAIN(GS_MAIN_ARGS_F)
#endif
{
  int i;
  time_t t;

  srand((unsigned) time(&t));
  /* Get the path to the directory containing the executable */
#if 0
  strncpy(Home, argv[0], BIGSTRINGLEN);
  if (strlen(Home) >= BIGSTRINGLEN)
  {
    printf("Weird problem with |%s|\n", Home);
    GSdelay(ERRORDELAY);
    exit(-1);
  }
  for(i=strlen(Home); (i>=0) && (Home[i]!='\\') && (Home[i]!='/'); i--)
    Home[i] = '\0';
#else
  Home[0] = 0;
#endif
  configure();
#if 0
  process_args ( argc , argv ) ;
#endif
  GSinterfaceinit("Mental Maps", InterfaceFile, Use_bstore, "w", G640x480x16, GS_MAIN_ARGS_A);

  if (!GSinitlog (Outputlogmode))
  {
    printf("Unable to initialize output.log\n");
    GSdelay(ERRORDELAY);
    exit(-1);
  }

  sprintf(VersionFile, "%s%s", Infopath, VERSION_INFO);
  GStogglecheck("Number", GSgetmenuitemname("Number", Num_cities/5 - 1)); /* Initialize number checkmark */
  for (i = 0; i < QUEST_NUM; i++)
    if (Question_item[i])
      GStogglecheck("Quest", GSgetmenuitemname("Quest", i));
  GSinterface();
  return(0);
}
