/**********************************************************************
All the Mental Maps module routines that don't belong anywhere else
(i.e., not simulation or datbase routines, nor routines called by the
interface handler).
**********************************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdarg.h>
#include <time.h>
#include "geosim.h"
#include "mmincl.h"
#include "triang.h"

/*--- tri.c ---*/
boolean triangle_transform(Point2, Point2, Point2,
                           Point2, Point2, Point2, double matrix[3][3]);

/*--- externally defined variables ---*/
extern Point2 Citypoints[72];
extern Point2 Userpoints[72];

/*--- Globals for warping ---*/
int      Currtriang; /*--- The index of the triangle last used ---*/

boolean  Transform = FALSE; /*--- FALSE => draw the map straight.
                                  TRUE => draw the warped map. ---*/
Edge     Edges[72 * 3 - 6], *tedge;
Triangle Triangles[72 * 2 - 4];
long     Numtriang;

/*------------------------------------------------------------------------------
  char_is_legal
    Returns TRUE if ch is a legal input character.
    Returns FALSE if ch is not a legal input character.
	make this a parameter that the app programmer specifies.
------------------------------------------------------------------------------*/
boolean char_is_legal (
  char  ch
)
{
  return ((isalpha(ch))
       || (ch == '.')
       || (ch == ',')
       || (ch == '-')
       || (ch == ' '))
  ;
}

/***********************************************************************
If -m option specified on command line, then mmap will dump a new
version of the .cty file for the current country.
***********************************************************************/
void dump_db_file(void)
{
  FILE      *cfp, *dfp;
  int       i, j;
  BigString cityname, pcode;
  BigString currfilename;
  BigString dumpfilename;
  BigString textline;
  int       x, y, low, hi;
  float     cost;
  long      pop;
  char      *p;

  sprintf(dumpfilename, "%s%s%s", Datapath, City[Curr_country].name,
          DUMPCITYSUFFIX);
  sprintf(currfilename, "%s%s%s", Datapath, City[Curr_country].name,
          CITYSUFFIX);
  if ((dfp = GSopen(dumpfilename, "w")) == NULL)
  {
    GSelog(0, "Unable to open dump city database file |%s|\n", dumpfilename);
    return;
  }
  if ((cfp = GSopen(currfilename, "r")) == NULL)
  {
    GSelog(0, "Unable to open city database file |%s|\n", currfilename);
    return;
  }
  do
  {
    fgets(textline, sizeof(textline), cfp);
    fputs(textline, dfp);
  }
  while (GSstricmp(textline, "tamrof\n") != 0);
  for (i = 0; i < City[Curr_country].num; i++)
  {
    fscanf(cfp, "%d", &x);
    fprintf(dfp, "%3d", City[Curr_country].clist[i].x -
            City[Curr_country].fudgex);
    fscanf(cfp, "%d", &y);
    fprintf(dfp, "%6d", City[Curr_country].clist[i].y -
            City[Curr_country].fudgey);
    if (City[Curr_country].pop_avail)
    {
      fscanf(cfp, "%ld", &pop);
      fprintf(dfp, "%12ld", pop);
    }
    if (City[Curr_country].low_avail)
    {
      fscanf(cfp, "%d", &low);
      fprintf(dfp, "%6d", low);
    }
    if (City[Curr_country].hi_avail)
    {
      fscanf(cfp, "%d", &hi);
      fprintf(dfp, "%6d", hi);
    }
    if (City[Curr_country].cost_avail)
    {
      fscanf(cfp, "%f", &cost);
      fprintf(dfp, "%8.1f", cost);
    }
    if (City[Curr_country].st_avail)
    {
/*      fscanf(cfp, "%s", pcode);     */
      strcpy(cityname, City[Curr_country].clist[i].name);
      p = cityname;
      while (*p != ',')
        p++;
      *p++ = '\0';
      sscanf(p, "%s", pcode);
      fprintf(dfp, "  %s", cityname);
      for (j = 0; j < CTYNAMEFIELDWIDTH - strlen(cityname); j++)
        fputc(' ', dfp);
      fprintf(dfp, "%s", pcode);
      for (j = 0; j < PCODEFIELDWIDTH - strlen(pcode); j++)
        fputc(' ', dfp);
    }
    else
    {
      fscanf(cfp, "%s", cityname);
      fprintf(dfp, "  %s", cityname);
      for (j = 0; j < CTYNAMEFIELDWIDTH - strlen(cityname); j++)
        fputc(' ', dfp);
    }
    fgets(textline, sizeof(textline), cfp);
    fprintf(dfp, "%3d", City[Curr_country].clist[i].labx -
            City[Curr_country].fudgex);
    fprintf(dfp, "%7d\n", City[Curr_country].clist[i].laby -
            City[Curr_country].fudgey);
  }
  fprintf(dfp, "%s\n", "endformat");
}


/***********************************************************************
Return TRUE iff the given point Q1 extended by a ray to the right
intersects the edge P1P2.
FALSE if the point is to the left of a horizontal line.
***********************************************************************/
boolean funny_intersect(
  Point2 P1,
  Point2 P2,
  Point2 Q1)
{

  if ((P1.y == P2.y) && (Q1.y == P1.y))
    return ((Q1.x >= MIN(P1.x, P2.x)) && (Q1.x <= MAX(P1.x, P2.x)));
  if ((Q1.y < MIN(P1.y, P2.y)) || (Q1.y > MAX(P1.y, P2.y)))
    return(FALSE);
  /* Now test if Q1 is left or right of P1P2 -- TRUE iff left */
  return(Q1.x <= (P1.x + ((P1.x - P2.x)/(P1.y - P2.y) * (Q1.y - P1.y))));
}

/***********************************************************************
Return TRUE iff the given point is in the triangle with the given index
***********************************************************************/
boolean in_triangle(
  Point2 P1,
  Point2 P2,
  Point2 P3,
  double x,
  double y)
{
  Point2    Q1;
  int       cnt = 0;

  Q1.x = x;
  Q1.y = y;
  if (funny_intersect(P1, P2, Q1)) cnt++;
  if (funny_intersect(P2, P3, Q1)) cnt++;
  if (funny_intersect(P3, P1, Q1)) cnt++;
  if ((cnt == 1) || (cnt == 3))
    return(TRUE);
  else
    return(FALSE);
}


/***********************************************************************
Warp a map point based on the triangle that it is in.
***********************************************************************/
void warp(
  double *currx,
  double *curry)
{
  int i;
  Point2 P1, P2, P3, Q1, Q2, Q3;
  Triangle *tp;
  static double matrix[3][3], temp;

  temp = Currtriang;
  if (Currtriang != -1)
  {
    tp = &Triangles[Currtriang];
    P1.x = tp->p[0]->x;
    P1.y = tp->p[0]->y;
    P2.x = tp->p[1]->x;
    P2.y = tp->p[1]->y;
    P3.x = tp->p[2]->x;
    P3.y = tp->p[2]->y;
  }
  if ((Currtriang == -1) || (!in_triangle(P1, P2, P3, *currx, *curry)))
  {
    for (i=0; i<Numtriang; i++)
    { /* First, find the triangle that contains the point */
      tp = &Triangles[i];
      P1.x = tp->p[0]->x;
      P1.y = tp->p[0]->y;
      P2.x = tp->p[1]->x;
      P2.y = tp->p[1]->y;
      P3.x = tp->p[2]->x;
      P3.y = tp->p[2]->y;
      if (in_triangle(P1, P2, P3, *currx, *curry))
      {
	Currtriang = i;
	break;
      }
    }
    /* Now, compute the warping */
    Q1.x = Triangles[i].p[0]->ux;
    Q1.y = Triangles[i].p[0]->uy;
    Q2.x = Triangles[i].p[1]->ux;
    Q2.y = Triangles[i].p[1]->uy;
    Q3.x = Triangles[i].p[2]->ux;
    Q3.y = Triangles[i].p[2]->uy;
    triangle_transform(P1, P2, P3, Q1, Q2, Q3, matrix);
  }
  /* now, run currx, curry through the transformation matrix */
  temp = *currx * matrix[0][0] + *curry * matrix[1][0] + matrix[2][0];
  *curry = *currx * matrix[0][1] + *curry * matrix[1][1] + matrix[2][1];
  *currx = temp;
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_login_win (
  void
)
{
  GSdologinscreen("Please type your name (First Last) and then hit Return:",
		  NO_BSTORE, login_return_func, GSconvertcolor("black"),
		  (char *)UserName);
  GSdisplayversion(VersionFile);
  if (QuitWinActive)
  {
    GSredrawnamedwindow("quit");
  }
  return;
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_root_win (
  void
)
{
  GSredrawnamedwindow("root");
  GSdisplayversion(VersionFile);
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_dump_results (
  void
)
{
  BigString msgstring;

  GSredrawnamedwindow("alert");
  GSsetcolor(GSconvertcolor("black"));
  GSsetfont(LARGE);
  if (DumpSuccess)
  {
    sprintf(msgstring, "The screen was saved to file:");
    GSwwritemsg(6, 30, msgstring);
    sprintf(msgstring, "   %s", DumpFileName);
    GSwwritemsg(6, 45, msgstring);
  }
  else
  {
    sprintf(msgstring, "An error occurred saving the screen.");
    GSwwritemsg(6, 15, msgstring);
    sprintf(msgstring, "No file was saved.");
    GSwwritemsg(6, 30, msgstring);
    sprintf(msgstring, "Check disk space and file permissions.");
    GSwwritemsg(6, 45, msgstring);
    GSbell();
  }
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_dump_win (
  void
)
{
  int       x, y;
  GSColor   color;

  GSredrawnamedwindow ("dump");
  highlightfield("dump", "filetype", GSconvertcolor("lgray"),
                 GSconvertcolor("black"), GSconvertcolor("white"));
  GSgetfieldrect("filetype", &x, &y, NULL, NULL, &color);
  GSsetcolor(color);
  GSsetfont(LARGE);
  switch (GraphFileFmt)
  {
    case BMP_16:   GSwwritenamedmsg(x+5, y+SMALL+2, "bmp16msg"); break;
    case GIF_16:   GSwwritenamedmsg(x+5, y+SMALL+2, "gif16msg"); break;
    case PS_GRAY:  GSwwritenamedmsg(x+5, y+SMALL+2, "psgraymsg"); break;
    case PS_COLOR: GSwwritenamedmsg(x+5, y+SMALL+2, "pscolormsg"); break;
    case NO_FILE:
    default:       GSwwritenamedmsg(x+5, y+SMALL+2, "nofilemsg"); break;
  }
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_alert_win (
  void
)
{
  switch(AlertNumber)
  {
    case 1: GSalert(2, Use_bstore, "aerr1", ""); break;
    case 2: GSalert(2, Use_bstore, "aerr2", ""); break;
    case 3: GSalert(1, Use_bstore, "nomapfile1", "nomapfile2", ""); break;
    default: break;
  }
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_quest (
  void
)
{
  int       x, y;
  GSColor   color;

  GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
  draw_city_circle(Citypoints[Citycnt-Levelsback].ux+x,
                   Citypoints[Citycnt-Levelsback].uy+y, color, NULL);
  switch (Prev_quest_status)
  {
    case TEMPERATURE:
      Slideval = &Answers[Citycnt-Levelsback].hi;
      askquest(Curr_city, "qtemph", FALSE, City[Curr_country].minhi,
               City[Curr_country].maxhi, " F", Answers[Citycnt-Levelsback].hi);
      break;
    case TEMP2:
      Slideval = &Answers[Citycnt-Levelsback].low;
      askquest(Curr_city, "qtempl", FALSE, City[Curr_country].minlow,
               City[Curr_country].maxlow, " F", Answers[Citycnt-Levelsback].low);
      break;
    case POPULATION:
      Slideval = &Answers[Citycnt-Levelsback].pop;
      askquest(Curr_city, "qpop", FALSE, City[Curr_country].minpop,
               City[Curr_country].maxpop, "", Answers[Citycnt-Levelsback].pop);
      break;
    case COST:
      Slideval = &Answers[Citycnt-Levelsback].cost;
      askquest(Curr_city, "qcost", FALSE, City[Curr_country].mincost,
               City[Curr_country].maxcost,
               " %", Answers[Citycnt-Levelsback].cost);
      break;
    case PREF:
      Slideval = &Answers[Citycnt-Levelsback].pref;
      askquest(Curr_city, "qpref", FALSE, 1, 10, "",
               Answers[Citycnt-Levelsback].pref);
      break;
  }
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_step (
  void
)
{
  int       x, y;
  GSColor   color;

  if (Show_results)
  {
    GSgetfieldrect("Question", &x, &y, NULL, NULL, &color);
    GSsetcolor(color);
    GSsetfont(LARGE);
    GSsetcurrentnamedwindow("cityname");
    GSwwritenamedmsg(x+10, y+SMALL, "resultsclue");
  }
  GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
  GSsetcurrentnamedwindow("map");
  draw_city_circle(Citypoints[StepCount-1].ux+x,
                   Citypoints[StepCount-1].uy+y, color, NULL);
  show_city(StepCount-1);
}

/*-------------------------------------------------------------------------------
-------------------------------------------------------------------------------*/
void redraw_cities (
  void
)
{
  int      i;
  int      x, y;
  GSColor  color;


  prompt_for_location();
  GSsetcurrentnamedwindow("map");
  GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
  for (i = 0 ; i < Citycnt-Levelsback ; i++)
    draw_city_circle(Citypoints[i].ux+x, Citypoints[i].uy+y, color, NULL);
}

/*-------------------------------------------------------------------------------
  Redraw all simulation information to the screen.  Typically done
  in response to clearing an overlapping window.
-------------------------------------------------------------------------------*/
boolean redraw_all(void)
{
  static boolean first = TRUE;

  if (first)
  {
    return (first = FALSE);
  }
  if (LoginInProcess == TRUE)
  {
    redraw_login_win();                 /* login */
    return TRUE;
  }
  redraw_root_win();                    /* root */
  GSredrawnamedwindow("cityname");      /* cityname */
  GSdisplayhelpmessage(CURR_MSG);       /* message */
  if (strcmp(MapFile, "") != 0)
  {
    draw_map(MapFile, FALSE);           /* map */
  }
  if (QuizInProgress)
  {
    redraw_cities();                    /* map and cityname */
  }
  if (!NotInStepThru)
  {
    redraw_step();                      /* map */
  }
  if (QuestionWinActive)
  {
    redraw_quest();                  /* question */
  }
  if (SummaryTableWinActive)
  {
    show_summary_table();               /* summary */
  }
  if (FinalMapActive)
  {
    show_final_map();                   /* map city name */
  }
  if (HelpWinActive)
  {
    GSdohelpscreens(Helpfile, CurrPage, Use_bstore, -1);
  }
  if (ResultWinActive)
  {
    show_results(ResultWinCity);
  }
  if (AlertNumber > 0)
  {
    redraw_alert_win();
  }
  if (DumpResultsActive)
  {
    redraw_dump_results();
  }
  if (QuitWinActive)
  {
    GSredrawnamedwindow("quit");
  }
  if (DumpWinActive)
  {
    redraw_dump_win();
  }
  return(TRUE);
}


/***********************************************************************
Local reader for a country map file.  Draw the map.
***********************************************************************/
boolean map_reader(
  BigString buffer,
  ReaderStatusType status,
  void *data)
{
  int fx, fy, fwd, fht;
  static int mapoffx, mapoffy, maphgt;
  GSColor fcolor;
  static GSColor linecolor;
  double currx, curry;
  static double lastx;
  static double lasty;
  int numkwds;
  FormatList format;
  BigString string;
  int i, numpoints;

  data = data; /*--- to quiet compiler ---*/
  if (status == READ_INIT)
  {
    lastx = lasty = -1.0;
    GSgetfieldrect("MapField", &mapoffx, &mapoffy, &fwd, &maphgt, &fcolor);
    if (Transform)
      GSgetfieldrect("Usercolor", &fx, &fy, &fwd, &fht, &linecolor);
    else
      GSgetfieldrect("Truecolor", &fx, &fy, &fwd, &fht, &linecolor);
    return(TRUE);
  }
  if (status == READ_FINAL)
    return(TRUE); /* Nothing to do */
  if (strcmp(buffer, "name=") == 0)
  {
    GSsetcurrentnamedwindow("cityname");
    GSgetfieldrect("CountryName", &fx, &fy, &fwd, &fht, &fcolor);
    GSsetcolor(fcolor);
    if (!GSrdphrase(string, BIGSTRINGLEN))
      return(FALSE);
    GSsetfont(LARGE);
    GSwwritemsg(fx, fy+SMALL, string);
    GSsetcurrentnamedwindow("Map");
  }
  else if (strcmp(buffer, "format=") == 0)
  { /* Process format statement */
    if (!GSrdformat(&numkwds, format))
      return(FALSE); /* GSrdformat should say why */
    if ((numkwds != 2) || (strcmp(format[0], "x") != 0) ||
			   (strcmp(format[1], "y") != 0))
      return (TRUE);  /* not the right format statement.  Skip this. */
    if (!GSrdword(string, BIGSTRINGLEN))
      return(FALSE);
    GSsetcolor(linecolor);
    while (strcmp(string, "endformat") != 0)
    {
      currx = MAPSCALE * atof(string);
      curry = MAPSCALE * GSrdreal();
      if (!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
      if (Transform)
        warp(&currx, &curry);
      if (lastx > 0)
      { /* Draw a line segment */
        GSwdrawline((int)currx+mapoffx, (int)curry+mapoffy,
                    (int)lastx+mapoffx, (int)lasty+mapoffy);
      }
      lastx = currx;
      lasty = curry;
    }
  } /* if format */
  else if (strcmp(buffer, "idrisipoints=") == 0)
  { /* Read in data for a map generated from IDRISI */
    if (!GSrdword(string, BIGSTRINGLEN))
      return(FALSE);
    while (strcmp(string, "endidrisi") != 0)
    { /* This section really is too trusting of the input.  CAS */
      numpoints = (int)GSrdinteger();
      lastx = GSrdreal();
      lasty = maphgt - GSrdreal();
      if (Transform)
        warp(&lastx, &lasty);
      GSsetcolor(linecolor);
      for (i=1; i<numpoints; i++)
      {
        currx = GSrdreal();
        curry = maphgt - GSrdreal();
        if (Transform)
          warp(&currx, &curry);
        GSwdrawline((int)currx+mapoffx, (int)curry+mapoffy,
                    (int)lastx+mapoffx, (int)lasty+mapoffy);
        lastx = currx;
        lasty = curry;
      }
      if (!GSrdword(string, BIGSTRINGLEN))
        return(FALSE);
    }
  }
  return(TRUE);
}


/***********************************************************************
Draw a country map.
***********************************************************************/
void draw_map(
  char *mapname,
  boolean warp)
{
  FILE *fp;

  Transform = warp;
  Currtriang = -1;
  if (!warp)
    GSredrawnamedwindow("map");
  if ((fp = GSopen(mapname, "rt")) == NULL)
  {
    GSelog(0, "Unable to open country map file |%s|\n", MapFile);
    strcpy(MapFile, ""); /* So redraw_all won't cycle */
    GSalert(2, Use_bstore, "aerr1", "");
    AlertNumber = 1;
    return;
  }
  if (!GSgenreader(fp, map_reader, "MM_ccoords", NULL))
  {
    GSelog(0, "Error processing country map file |%s|\n", MapFile);
    strcpy(MapFile, ""); /* So redraw_all won't cycle */
    GSalert(2, Use_bstore, "aerr2", "");
    AlertNumber = 2;
    return;
  }
  fclose(fp);
  return;
}


/***********************************************************************
Put up the dialog box and slider for a question.
***********************************************************************/
void askquest (
  int     city,
  char    *qstring,
  boolean label,
  long    cmin,
  long    cmax,
  char    *units,
  long    prevval
)
{
  int x, y, wd, ht;
  GSColor color;
  BigString string;
  int offset;

/*  GSdisablebutton("cityname", "goback", DRAW);
  GSdisablebutton("cityname", "start", DRAW);*/
  GSsetcurrentnamedwindow("cityname");
  GSgetfieldrect("questionbackground", &x, &y, &wd, &ht, &color);
  GSwclearrect(x, y, wd, ht);
  GSsetcurrentnamedwindow("question");
  if (QuestionWinActive)
    GSredrawnamedwindow("question");
  else
    GSdrawnamedwindow("question", Use_bstore, POPUP);
  QuestionWinActive = TRUE;
  GSgetfieldrect("Questfield", &x, &y, &wd, &ht, &color);
  GSsetcolor(color);
  GSsetfont(SMALL);
  GSwwritenamedmsg(x, y+SMALL, qstring);
  GSgetfieldrect("Namefield", &x, &y, &wd, &ht, &color);
  GSsetcolor(color);
  GSwwritemsg(x, y+SMALL, City[Curr_country].clist[city].name);
  /* Set up the slider bar */
  GSgetfieldrect("Slidemin", &x, &y, &wd, &ht, &color);
  Currslide.min = cmin;
  Currslide.range = cmax - cmin;
  strcpy(Units, units);
  if (label)
    GSwwritemsg(x, y+SMALL, "No");
  else
  {
    sprintf(string, "%ld%s", cmin, Units);
    GSwwritemsg(x, y+SMALL, string);
  }
  GSgetfieldrect("Slidemax", &x, &y, &wd, &ht, &color);
  if (label)
    GSwwritemsg(x, y+SMALL, "Yes");
  else {
    sprintf(string, "%ld%s", cmax, Units);
    GSwwritemsg(x, y+SMALL, string);
  }
  GSgetfieldrect("Questslidetick", &Currslide.x, &Currslide.y, &Currslide.wd,
		 &Currslide.ht, &Currslide.tickcolor);
  GSgetfieldrect("Questslide", &Currslide.x, &Currslide.y, &Currslide.wd,
		 &Currslide.ht, &Currslide.barcolor);
  Currslide.writeval = !label;
  if (prevval >= 0) /* draw slider and value that user entered before */
  {
    sprintf(string, "%3ld%s", prevval, Units);
    offset = (Currslide.wd - GStextwidth(string))/2;
    GSsetcolor(Currslide.tickcolor);
    GSwwritemsg(Currslide.x+offset, Currslide.y+SLVALOFFY+SMALL, string);
    (void)drawslider(Currslide.x, Currslide.y, Currslide.wd, Currslide.ht,
       Currslide.barcolor, Currslide.tickcolor,
       Currslide.x + ((float)(prevval - Currslide.min)
       / (float)Currslide.range) * Currslide.wd);
    GSdisplayhelpmessage(21);
  }
  else
  {
    strcpy(string, "Move slidebar.");
    offset = (Currslide.wd - GStextwidth(string))/2;
    GSsetcolor(Currslide.tickcolor);
    GSwwritemsg(Currslide.x+offset, Currslide.y+SLVALOFFY+SMALL, string);
    (void)drawslider(Currslide.x, Currslide.y, Currslide.wd, Currslide.ht,
       Currslide.barcolor, Currslide.tickcolor, Currslide.x + Currslide.wd/2);
    GSdisplayhelpmessage(17);
  }
}


/***********************************************************************
Deliver the quiz questions for a city.  First randomly select the
next city, then start the process of asking the questions.  All
interface functions will direct control back to this function.
***********************************************************************/
void askcity(void)
{
  int mfx, mfy, mfwd, mfht;
  int x, y;
  GSColor mfcolor;
  GSColor color;
  MyRect currlabel;

  /* We actually ask LOCATION first, but the switch is set up to allow
     other questions to fall-through if they are not asked. */
  if (Citycnt < 0)
    Quest_status = LOCATION;
  switch (Quest_status) {
    case TEMPERATURE:
      Prev_quest_status = Quest_status;
      Quest_status = TEMP2;
      if (Question_item[TEMPERATURE] && City[Curr_country].hi_avail)
      {
        Slideval = &Answers[Citycnt-Levelsback].hi;
        askquest(Curr_city, "qtemph", FALSE,
                 City[Curr_country].minhi, City[Curr_country].maxhi, " F",
                 Answers[Citycnt-Levelsback].hi);
        break;
      }
    case TEMP2:
      Prev_quest_status = Quest_status;
      Quest_status = POPULATION;
      if (Question_item[TEMPERATURE] && City[Curr_country].low_avail)
      {
        Slideval = &Answers[Citycnt-Levelsback].low;
        askquest(Curr_city, "qtempl", FALSE,
                 City[Curr_country].minlow, City[Curr_country].maxlow,
                 " F", Answers[Citycnt-Levelsback].low);
        break;
      }
    case POPULATION:
      Prev_quest_status = Quest_status;
      Quest_status = COST;
      if (Question_item[POPULATION] && City[Curr_country].pop_avail)
      {
        Slideval = &Answers[Citycnt-Levelsback].pop;
        askquest(Curr_city, "qpop", FALSE,
                 City[Curr_country].minpop, City[Curr_country].maxpop, "",
                 Answers[Citycnt-Levelsback].pop);
        break;
      }
    case COST:
      Prev_quest_status = Quest_status;
      Quest_status = PREF;
      if (Question_item[COST] && City[Curr_country].cost_avail)
      {
        Slideval = &Answers[Citycnt-Levelsback].cost;
        askquest(Curr_city, "qcost", FALSE,
                 City[Curr_country].mincost, City[Curr_country].maxcost,
                 " %", Answers[Citycnt-Levelsback].cost);
        break;
      }
    case PREF:
      Prev_quest_status = Quest_status;
      Quest_status = LOCATION;
      if (Question_item[PREF])
      {
        Slideval = &Answers[Citycnt-Levelsback].pref;
        askquest(Curr_city, "qpref", FALSE, 1, 10, "",
        Answers[Citycnt-Levelsback].pref);
        break;
      }
    case LOCATION:
if (Citycnt >= 0)
{
  GSremovenamedwindow("Question"); 
  QuestionWinActive = FALSE;
  if (!Use_bstore)
  {
    redraw_all();
    GSsetcurrentnamedwindow("map");
    GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
    draw_city_circle(Citypoints[Citycnt-Levelsback].ux+x,
                     Citypoints[Citycnt-Levelsback].uy+y, color, NULL);
  }
}
      if (Citycnt >= Levelsback)
        GSenablebutton("cityname", "GoBack", DRAW);

      if (Levelsback > 0)
      {
        Levelsback--;
        Curr_city = (int)Answers[Citycnt-Levelsback].citynum;
      }
      else
      {
        /* Pick the next city at random */
        Citycnt++;
        if ((Citycnt % 5 == 0) && (Citycnt >= 5)) /* otherwise you're trying to disable
        											the -1'th menu item - JBR */
          GSdisablemenuitem("number", GSgetmenuitemname("number", (Citycnt / 5) - 1));
        if ((Citycnt == Num_cities) || (Citycnt == City[Curr_country].num))
        {
          finish_quiz(); /* Then we are all done. */
          return;
        }
        if (Num_cities != 68)
        {
          Curr_city = rand() % City[Curr_country].num;
          GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
          currlabel.x = City[Curr_country].clist[Curr_city].labx+mfx;
          currlabel.y = City[Curr_country].clist[Curr_city].laby+mfy;
          currlabel.wd = GStextwidth(City[Curr_country].clist[Curr_city].name);
          currlabel.ht = GStextheight(City[Curr_country].clist[Curr_city].name);
          while (City[Curr_country].clist[Curr_city].mark ||
                 !clearforlabel(currlabel))
          {
            Curr_city = rand() % City[Curr_country].num;
            currlabel.x = City[Curr_country].clist[Curr_city].labx+mfx;
            currlabel.y = City[Curr_country].clist[Curr_city].laby+mfy;
            currlabel.wd = GStextwidth(City[Curr_country].clist[Curr_city].name);
            currlabel.ht = GStextheight(City[Curr_country].clist[Curr_city].name);
          }
        }
        else
        {
          while (City[Curr_country].clist[Curr_city].mark)
          {
            Curr_city = rand() % City[Curr_country].num;
          }
        }
        City[Curr_country].clist[Curr_city].mark = TRUE;
        Citypoints[Citycnt].x = (double)City[Curr_country].clist[Curr_city].x;
        Citypoints[Citycnt].y = (double)City[Curr_country].clist[Curr_city].y;
        Answers[Citycnt].citynum = Curr_city;
        /* DEBUG */
        /*GSelog(0, "%s %5.0lf %5.0lf \n",
               City[Curr_country].clist[Curr_city].name,
        Citypoints[Citycnt].x, Citypoints[Citycnt].y);*/
      }
      /* Then put up the question about location */
      prompt_for_location();
      GSdisplayhelpmessage(Savemsg = 6);
      Quest_status = TEMPERATURE;
      GSsetcurrentnamedwindow("Map");
      break;
    case RELOCATION:
      if (!Questesc) /*--- go back from city name window, not quest window ---*/
        Levelsback++;
      if (Levelsback == Citycnt)
        GSdisablebutton("cityname", "GoBack", DRAW);

      Curr_city = (int)Answers[Citycnt-Levelsback].citynum;
      GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
      GSsetcurrentnamedwindow("Map");
      erase_city_circle(Citypoints[Citycnt-Levelsback].ux+mfx,
                        Citypoints[Citycnt-Levelsback].uy+mfy,
                        Citycirclepixels[Citycnt-Levelsback]);
      /* Then put up the question about location */
      prompt_for_location();
      GSdisplayhelpmessage(Savemsg = 6);
      Quest_status = TEMPERATURE;
      if (Questesc)
      {
        Questesc = FALSE;
      }
      break;
    default:  break;
  } /* switch */
}


/***********************************************************************
Prompt the user to locate a city.
The prompt appears in the CityName window.
The two-digit postal code stored in the database is converted
to a full state name for the prompt.
***********************************************************************/
void prompt_for_location(void)
{
  int x, y, wd, ht;
  GSColor color;
  char *p;
  BigString name;
  BigString string;

  GSsetcurrentnamedwindow("cityname");
  GSgetfieldrect("Question", &x, &y, &wd, &ht, &color);
  GSwclearrect(x, y, wd, ht);
  GSsetcolor(color);
  strcpy(name, City[Curr_country].clist[Curr_city].name);
  p = name;                             /* Remove state postal code */
  while (*p++);
  p-=3;
  *p = '\0';
  strcpy(name, fullname(City[Curr_country].clist[Curr_city].name));
  sprintf(string, "%s%s", "Click on ", name);
  GSsetfont(LARGE);
  GSwwritemsg(x+10, y+SMALL, string);
}


/***********************************************************************
Convert a two digit postal code for a state to that state's full name.
***********************************************************************/
char *fullname
 (char *name)
{
  char **pp;
  char *p;
  BigString buffer;
  BigString pcode;

  strcpy(buffer, name);
  p = buffer;                             /* Remove state postal code */
  while (*p++);
  p-=3;
  strcpy(pcode, p);
  *p = '\0';

  pp = (char **)StatePCodeTable;
  while (*pp != NULL)
  {
    if (strstr(pcode, *pp))
    {
      pp++;
      break;
    }
    pp+=2;
  }
  if (*pp == NULL)
  {
    return(name);
  }
  else
  {
    return (strcat(buffer, *pp));
  }
}


/***********************************************************************
Returns TRUE if "label" does NOT overlap any other labels or any
city markers on the map.
***********************************************************************/
boolean clearforlabel(
  MyRect label)
{
  MyRect label_i, city_circle_rect;
  int i;
  int mfx, mfy, mfwd, mfht;
  GSColor mfcolor;

  GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
  for (i = 0; i < Num_cities; i++)
  {
    label_i.x = City[Curr_country].clist[(int)Answers[i].citynum].labx+mfx;
    label_i.y = City[Curr_country].clist[(int)Answers[i].citynum].laby+mfy;
    label_i.wd =
      GStextwidth(City[Curr_country].clist[(int)Answers[i].citynum].name);
    label_i.ht =
      GStextheight(City[Curr_country].clist[(int)Answers[i].citynum].name);
    if (rects_overlap(label, label_i))
      return FALSE;
    city_circle_rect.x = Citypoints[i].x+mfx-2;
    city_circle_rect.y = Citypoints[i].y+mfy-2;
    city_circle_rect.wd = 4;
    city_circle_rect.ht = 4;
    if (rects_overlap(label, city_circle_rect))
      return FALSE;
  }
  return TRUE;
}


/***********************************************************************
Record positon of label for 'city' in 'Citylabels'.
Argument passed to 'pos' determines where in relation to city location
  the label should go
NOT IN USE, BUT KEEP FOR POSSIBLE FUTURE USE
***********************************************************************/
/* void setlabelpos(
  int city,
  int pos)
{
  int mfx, mfy, mfwd, mfht;
  GSColor mfcolor;

  GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);
  GSsetfont(SMALL);
  Citylabels[city].wd =
    GStextwidth(City[Curr_country].clist[(int)Answers[city].citynum].name);
  Citylabels[city].ht =
    GStextheight(City[Curr_country].clist[(int)Answers[city].citynum].name);
  switch(pos) {
    case 0: / to right of city circle /
      Citylabels[city].x = Citypoints[city].x+mfx+4;
      Citylabels[city].y = Citypoints[city].y+mfy+3;
      break;
    case 1: / above city circle /
      Citylabels[city].x = Citypoints[city].x+mfx-Citylabels[city].wd/2;
      Citylabels[city].y = Citypoints[city].y+mfy-(Citylabels[city].ht/2);
      break;
    case 2: / to left of city circle /
      Citylabels[city].x = Citypoints[city].x+mfx-(Citylabels[city].wd+4);
      Citylabels[city].y = Citypoints[city].y+mfy+3;
      break;
    case 3: / under city circle /
      Citylabels[city].x = Citypoints[city].x+mfx-Citylabels[city].wd/2;
      Citylabels[city].y = Citypoints[city].y+mfy+6+Citylabels[city].ht/2;
      break;
  }
} */


/***********************************************************************
Test to see if two labels (1 and 2) intersect.
NOT IN USE, BUT KEEP FOR POSSIBLE FUTURE USE
***********************************************************************/
boolean rects_overlap(
  MyRect rect1,
  MyRect rect2)
{
  if ((rect1.x > rect2.x)
   && (rect1.x < (rect2.x + rect2.wd))
   && (rect1.y < rect2.y)
   && (rect1.y > (rect2.y - rect2.ht)))
    return (TRUE); /* bottom left corner of 1 is inside 2 */

  if ((rect1.x > rect2.x)
   && (rect1.x < (rect2.x + rect2.wd))
   && (rect1.y - rect1.ht < rect2.y)
   && (rect1.y - rect1.ht > (rect2.y - rect2.ht)))
    return (TRUE); /* top left corner of 1 is inside 2 */

  if ((rect1.x + rect1.wd > rect2.x)
   && (rect1.x + rect1.wd < (rect2.x + rect2.wd))
   && (rect1.y < rect2.y)
   && (rect1.y > (rect2.y - rect2.ht)))
    return (TRUE); /* bottom right corner of 1 is inside 2 */

  if ((rect1.x + rect1.wd > rect2.x)
   && (rect1.x + rect1.wd < (rect2.x + rect2.wd))
   && (rect1.y - rect1.ht < rect2.y)
   && (rect1.y - rect1.ht > (rect2.y - rect2.ht)))
    return (TRUE); /* top right corner of 1 is inside 2 */

  if ((rect2.x > rect1.x)
   && (rect2.x < (rect1.x + rect1.wd))
   && (rect2.y < rect1.y)
   && (rect2.y > (rect1.y - rect1.ht)))
    return (TRUE); /* bottom left corner of 2 is inside 1 */

  if ((rect2.x > rect1.x)
   && (rect2.x < (rect1.x + rect1.wd))
   && (rect2.y - rect2.ht < rect1.y)
   && (rect2.y - rect2.ht > (rect1.y - rect1.ht)))
    return (TRUE); /* top left corner of 2 is inside 1 */

  if ((rect2.x + rect2.wd > rect1.x)
   && (rect2.x + rect2.wd < (rect1.x + rect1.wd))
   && (rect2.y < rect1.y)
   && (rect2.y > (rect1.y - rect1.ht)))
    return (TRUE); /* bottom right corner of 2 is inside 1 */

  if ((rect2.x + rect2.wd > rect1.x)
   && (rect2.x + rect2.wd < (rect1.x + rect1.wd))
   && (rect2.y - rect2.ht < rect1.y)
   && (rect2.y - rect2.ht > (rect1.y - rect1.ht)))
    return (TRUE); /* top right corner of 2 is inside 1 */

  return (FALSE); /* if it gets this far, no overlap */
}


/***********************************************************************
TRUE if currpos for label of currcity does not cover anything and does
  not run off of map window
NOT IN USE, BUT KEEP FOR POSSIBLE FUTURE USE
***********************************************************************/
/* boolean posOK(
  int currcity,
  int currpos)
{
  int i;
  MyRect city_circle_rect;
  int mfx, mfy, mfwd, mfht;
  GSColor mfcolor;

  setlabelpos(currcity, currpos);
  GSgetfieldrect("MapField", &mfx, &mfy, &mfwd, &mfht, &mfcolor);

  / Does label run off right side of map window at currpos? /
  if (Citylabels[currcity].x + Citylabels[currcity].wd + 5 > mfwd)
    return FALSE;

  / Does label run off left side of map window at currpos? /
  if (Citylabels[currcity].x < mfx)
    return FALSE;

  / Does label overlap with another label at currpos? /
  for (i = 0 ; i < currcity ; i++)
  {
    if (rects_overlap(Citylabels[currcity], Citylabels[i]))
      return FALSE;
  }

  / Does label overlap with any city circles at currpos? /
  for (i = 0 ; i < Num_cities ; i++)
  {
    / Check actual location for city i /
    city_circle_rect.x = Citypoints[i].x+mfx-2;
    city_circle_rect.y = Citypoints[i].y+mfy-2;
    city_circle_rect.wd = 4;
    city_circle_rect.ht = 4;
    if (rects_overlap(Citylabels[currcity], city_circle_rect))
      return FALSE;

    / Check user location for city i /
    city_circle_rect.x = Citypoints[i].ux+mfx-2;
    city_circle_rect.y = Citypoints[i].uy+mfy-2;
    if (rects_overlap(Citylabels[currcity], city_circle_rect))
      return FALSE; 
  }
  return TRUE; / if we got thru all tests, currpos is OK /
} */

/***********************************************************************
Recursive function for placing city labels.
NOT IN USE, BUT KEEP FOR POSSIBLE FUTURE USE
***********************************************************************/
/* boolean placelabels(
  int currcity,
  int currpos)
{
  //GSelog(0, "city: %d     pos: %d\n", currcity, currpos);
  if (currcity == Num_cities)
    return (TRUE);                      / all labels placed without overlap /
  else if (currpos == 4)
    return (FALSE);                     / backtrack /
  else if (posOK(currcity, currpos))
    if (placelabels(currcity + 1, 0))
      return TRUE;
    else
      return (placelabels(currcity, currpos + 1));
  else
    return (placelabels(currcity, currpos + 1));   / try next position /
} */


/***********************************************************************
Show the real location and user placement of 'city'.
***********************************************************************/
void show_city(
  int city)
{
  int x, y, wd, ht;
  GSColor color;

  GSsetcurrentnamedwindow("map");
  GSgetfieldrect("Mapfield", &x, &y, &wd, &ht, &color);
  draw_city_circle(Citypoints[city].x+x, Citypoints[city].y+y, color, NULL);
  GSsetcolor(color);
  GSsetfont(SMALL);
  GSwwritemsg(City[Curr_country].clist[(int)Answers[city].citynum].labx+x,
              City[Curr_country].clist[(int)Answers[city].citynum].laby+y,
              City[Curr_country].clist[(int)Answers[city].citynum].name);
  GSgetfieldrect("Connector", &x, &y, &wd, &ht, &color);
  GSsetcolor(color);
  if (!Moving)
    GSwdrawline((int)Citypoints[city].x+x, (int)Citypoints[city].y+y,
		(int)Citypoints[city].ux+x, (int)Citypoints[city].uy+y);

}

/***********************************************************************
Write the user's answers to output.log
***********************************************************************/
void log_answers(void)
{
  time_t timevar;
  int i;

  timevar = time(NULL);
  if (strlen(UserName) > 0)
    GSolog("Current user is %s\n", UserName);
  GSolog("Current date is %s", ctime(&timevar));
  GSolog("Current country is %s\n", City[Curr_country].name);
  GSolog("Number of cities is %d\n", Num_cities);
  for (i=0; i<Num_cities; i++)
  {
    GSolog("%s: %ld %ld %ld %ld %ld %ld %ld\n",
           City[Curr_country].clist[(int)Answers[i].citynum].name,
           (long)Citypoints[i].ux, (long)Citypoints[i].uy,
           Answers[i].hi, Answers[i].low, Answers[i].pop, Answers[i].cost,
           Answers[i].pref);
  }
  GSolog("\n\n");
}


/***********************************************************************
Show all city locations and warped map
***********************************************************************/
void show_final_map(void)
{
  int x, y, i;
  GSColor color;

  GSgetfieldrect("Usercolor", &x, &y, NULL, NULL, &color);
  for (i = 0 ; i < Num_cities ; i++)
  {
    show_city(i);
    draw_city_circle(Citypoints[i].ux+x, Citypoints[i].uy+y, color, NULL);
  }
  GSgetfieldrect("Question", &x, &y, NULL, NULL, &color);
  GSsetcolor(color);
  GSsetfont(LARGE);
  GSsetcurrentnamedwindow("cityname");
  GSwwritenamedmsg(x+10, y+SMALL, "quizendquest");
  if (!Moving)
    draw_warped_map();
  FinalMapActive = TRUE;
}


/***********************************************************************
Show real and user locations for one city at a time.
***********************************************************************/
void step_thru(void)
{
  int x, y, wd, ht;
  GSColor color;

  GSsetcurrentnamedwindow("map");
  GSvisiblebutton("Message", "go_on", DRAW);
  draw_map(MapFile, FALSE);
  GSgetfieldrect("Usercolor", &x, &y, &wd, &ht, &color);
  if (StepCount < Num_cities)
  {
    draw_city_circle(Citypoints[StepCount].ux+x,
                   Citypoints[StepCount].uy+y, color, NULL);
    show_city(StepCount);
    StepCount++;
  }
  else
  {
    GSinvisiblebutton("Message", "go_on", DRAW);
    NotInStepThru = TRUE;
    if (Show_results)
      show_summary_table();
    else
    {
      show_final_map();
      if (Show_results)    
        GSdisplayhelpmessage(Savemsg = 8);
      else
        GSdisplayhelpmessage(Savemsg = 13);
      GSenablebutton("root", "number", DRAW);
      GSenablebutton("root", "quest", DRAW);
      GSenablebutton("root", "country", DRAW);
    }
  }
}


/***********************************************************************
Show summary table of user's answers compared with correct answers.
***********************************************************************/
void show_summary_table(void)
{
  int x, y;
  int dumy, wd, ht;
  int i;
  GSColor color;
  BigString astring;
  int columnnum = 1;

  GSsetcurrentnamedwindow("cityname");
  GSgetfieldrect("Question", &x, &y, &wd, &ht, &color);
  GSwclearrect(x, y, wd, ht);
  GSenablebutton("SummaryTable", "summtabdone", NO_DRAW);
/*  GSdisablebutton("cityname", "goback", DRAW);
  GSdisablebutton("cityname", "start", DRAW);*/
  if (!SummaryTableWinActive)
    GSdrawnamedwindow("SummaryTable", Use_bstore, POPUP);
  else
    GSredrawnamedwindow("SummaryTable");
  SummaryTableWinActive = TRUE;
  GSsetcurrentnamedwindow("SummaryTable");
  GSsetfont(LARGE);
  GSdisplayhelpmessage(22);
  y = 19 + SMALL;
  GSgetfieldrect("citynamecol", &x, &dumy, &wd, &ht, &color);
  GSsetcolor(color);
  GSsetfont(SMALL);
  GSwwritemsg(x+4, y-14, "City");
  if (Question_item[TEMPERATURE] && City[Curr_country].hi_avail)
  {
    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
    sprintf(astring, "%s", "Correct");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, y+5, x, ht);
    sprintf(astring, "%s", "Yours");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s", "High Temp");
    GSwwritemsg(x-GStextwidth(astring)/2, y-14, astring);
  }
  if (Question_item[TEMPERATURE] && City[Curr_country].low_avail)
  {
    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
    sprintf(astring, "%s", "Correct");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, y+5, x, ht);
    sprintf(astring, "%s", "Yours");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s", "Low Temp");
    GSwwritemsg(x-GStextwidth(astring)/2, y-14, astring);
  }
  if (Question_item[POPULATION] && City[Curr_country].pop_avail)
  {
    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
    sprintf(astring, "%s", "Correct");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, y+5, x, ht);
    sprintf(astring, "%s", "Yours");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s", "Population");
    GSwwritemsg(x-GStextwidth(astring)/2, y-14, astring);
  }
  if (Question_item[COST] && City[Curr_country].cost_avail)
  {
    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
    sprintf(astring, "%s", "Correct");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, y+5, x, ht);
    sprintf(astring, "%s", "Yours");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);

    sprintf(astring, "%s", "Cost of Living");
    GSwwritemsg(x-GStextwidth(astring)/2, y-14, astring);
  }
  if (Question_item[PREF])
  {
    sprintf(astring, "%s%d", "column", columnnum);
    columnnum++;
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
    sprintf(astring, "%s", "Your");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2-2, y-14, astring);
    sprintf(astring, "%s", "Rating");
    GSwwritemsg(x+wd/2-GStextwidth(astring)/2-2, y, astring);
  }
  GSgetfieldrect("bottom", &x, &dumy, &wd, &ht, &color);
  GSwdrawline(0, dumy, wd-1, dumy);
  if (columnnum < 10)                   /* draw line at right end of last
                                           column */
  {
    sprintf(astring, "%s%d", "column", columnnum);
    GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
    GSwdrawline(x, 0, x, ht);
  }
  y += 5;
  GSgetfieldrect("whole", &x, &dumy, &wd, &ht, &color);
  GSwdrawline(0, y, wd-1, y);
  for (i = 0 ; i < Num_cities ; i++)
  {
    columnnum = 1;
    y += 14;
    GSgetfieldrect("citynamecol", &x, &dumy, &wd, &ht, &color);
    GSwwritemsg(x+4, y,
                City[Curr_country].clist[(int)Answers[i].citynum].name);
    if (Question_item[TEMPERATURE] && City[Curr_country].hi_avail)
    {
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      sprintf(astring, "%d",
              City[Curr_country].clist[(int)Answers[i].citynum].hi);
      GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      if (Answers[i].hi > 0)
      {
        sprintf(astring, "%ld", Answers[i].hi);
        GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      }
      else
      {
        sprintf(astring, "%s", "none");
        GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);
      }
    }
    if (Question_item[TEMPERATURE] && City[Curr_country].low_avail)
    {
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      sprintf(astring, "%d",
              City[Curr_country].clist[(int)Answers[i].citynum].low);
      GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      if (Answers[i].low > 0)
      {
        sprintf(astring, "%ld", Answers[i].low);
        GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      }
      else
      {
        sprintf(astring, "%s", "none");
        GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);
      }
    }
    if (Question_item[POPULATION] && City[Curr_country].pop_avail)
    {
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      sprintf(astring, "%ld",
              City[Curr_country].clist[(int)Answers[i].citynum].pop);
      GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("9999999")/2,
                  y, astring);
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      if (Answers[i].pop > 0)
      {
        sprintf(astring, "%ld", Answers[i].pop);
        GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("9999999")/2,
                  y, astring);
      }
      else
      {
        sprintf(astring, "%s", "none");
        GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);
      }
    }
    if (Question_item[COST] && City[Curr_country].cost_avail)
    {
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      sprintf(astring, "%ld",
              City[Curr_country].clist[(int)Answers[i].citynum].cost);
      GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      if (Answers[i].cost > 0)
      {
        sprintf(astring, "%ld", Answers[i].cost);
        GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("99")/2, y, astring);
      }
      else
      {
        sprintf(astring, "%s", "none");
        GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);
      }
    }
    if (Question_item[PREF])
    {
      sprintf(astring, "%s%d", "column", columnnum);
      columnnum++;
      GSgetfieldrect(astring, &x, &dumy, &wd, &ht, &color);
      if (Answers[i].pref > 0)
      {
        sprintf(astring, "%ld", Answers[i].pref);
        GSwwritemsg(x+wd/2-GStextwidth(astring)+GStextwidth("9")/2, y, astring);
      }
      else
      {
        sprintf(astring, "%s", "none");
        GSwwritemsg(x+wd/2-GStextwidth(astring)/2, y, astring);
      }
    }
  }
}


/***********************************************************************
Draw the warped map based on user placement of cities.
***********************************************************************/
void draw_warped_map(void)
{
  int i;
  long edgecnt;
  int qx, qy, qwd, qht;
  GSColor qcolor;

  GSgetfieldrect("Mapfield", &qx, &qy, &qwd, &qht, &qcolor);
  i = Num_cities;
  Citypoints[i].x = Citypoints[i].y = Citypoints[i+1].y = Citypoints[i+2].x
    = 0.0;
  Citypoints[i+1].x = Citypoints[i+3].x = qwd;
  Citypoints[i+2].y = Citypoints[i+3].y = qht;
  Citypoints[i].ux = Citypoints[i].uy = Citypoints[i+1].uy = Citypoints[i+2].ux
    = 0.0;
  Citypoints[i+1].ux = Citypoints[i+3].ux = qwd;
  Citypoints[i+2].uy = Citypoints[i+3].uy = qht;
  edgecnt = 0;
  Numtriang = triangulate(Citypoints, (long)(4+Num_cities),
                          Edges, &edgecnt, Triangles);
  if (Numtriang < 0)
    GSelog(-1, "Problem with triangulation: %ld\n", Numtriang);
  else if (GSDebug)
  { /* Draw triangulation. */
    GSsetcurrentnamedwindow("map");
    GSsetcolor(qcolor);
    for (i=0; i<edgecnt; i++)
    {
      tedge = &(Edges[i]);
      GSwdrawline((int)tedge->p->x+qx, (int)tedge->p->y+qy,
		  (int)tedge->q->x+qx, (int)tedge->q->y+qy);
    }
  }
  GSsetlinesize(3);
  draw_map(WarpMapFile, TRUE);
  GSsetlinesize(1);
}

/***********************************************************************
Run thru cities one-by-one, allowing user to move them. Any city which
has been seen can be moved.
***********************************************************************/
boolean do_moving(void)
{
  int x, y, wd, ht;
  GSColor color;

  Move_labels = TRUE;
  GSsetcurrentnamedwindow("map");
  GSgetfieldrect("Mapfield", &x, &y, &wd, &ht, &color);
  draw_city_circle(City[Curr_country].clist[Move_index].x+x,
                   City[Curr_country].clist[Move_index].y+y, color, NULL);
  GSsetcolor(color);
  GSsetfont(SMALL);
  GSwwritemsg(City[Curr_country].clist[Move_index].labx+x,
                  City[Curr_country].clist[Move_index].laby+y,
                  City[Curr_country].clist[Move_index].name);
  GSdisplayhelpmessage(24);
  GSenabledragarea("Map", "Dragmap");
  Move_index++;
  if (Move_index == City[Curr_country].num)
  {
    GSdisablebutton("Message", "next", DRAW);
    GSdisplayhelpmessage(25);
  }
  return (TRUE);
}

/***********************************************************************
All quiz questions have been asked, now begin finish-up procedure.
***********************************************************************/
void finish_quiz(void)
{
  int qx, qy, qwd, qht;
  GSColor qcolor;
  int i;
  BigString longline;
  int dum;
  FILE *fp;
  int x, y, wd, ht;
  GSColor color;

  QuizInProgress = FALSE;
  /*--- if Moving, then dispense with all the regular stuff ---*/
  if (Moving)
  {
    Move_index = 0;
    GSvisiblebutton("message", "next", DRAW);
    do_moving();
    return;
  }
  for (i = 0; i < 4; i++)
    GSenablemenuitem("number", GSgetmenuitemname("number", i));
  GSdisabledragarea("Map", "Dragmap");
  GSdisablebutton("cityname", "GoBack", DRAW);
  if (ReadDebugInput)
  {
    if ((fp = GSopen("input.txt", "r")) == NULL)
      GSelog(-1, "Unable to open input.txt\n");
    fgets(longline, BIGSTRINGLEN, fp);
    sscanf(longline, "%d", &Num_cities);
    for (i=0; i<Num_cities; i++)
    {
      if(fgets(longline, BIGSTRINGLEN, fp) == NULL)
        GSelog(-1, "Error reading input.txt\n");
      sscanf(longline, "%d %lf %lf %lf %lf\n", &dum,
             &Citypoints[i].x, &Citypoints[i].y,
	     &Citypoints[i].ux, &Citypoints[i].uy);
    }
    fclose(fp);
  }
  GSgetfieldrect("Question", &qx, &qy, &qwd, &qht, &qcolor);
  GSsetcolor(qcolor);
  GSwclearrect(qx, qy, qwd, qht);
  GSsetfont(SMALL);
  GSsetcurrentnamedwindow("Map");
  if ((Question_item[TEMPERATURE] && City[Curr_country].hi_avail)
  ||  (Question_item[TEMPERATURE] && City[Curr_country].low_avail)
  ||  (Question_item[POPULATION] && City[Curr_country].pop_avail)
  ||  (Question_item[COST] && City[Curr_country].cost_avail)
  ||  (Question_item[PREF]))
  {                                     /* Allow user to pop up results to
                                           questions */
    GSenabledragarea("Map", "Dragmap");
    Show_results = TRUE;
  }
  if (Moving)
  {
    Move_labels = TRUE;
    GSenabledragarea("Map", "Dragmap");
  }
  log_answers();                        /* Write user's answers to log file */
  if (StepThru && !Moving)
  {
    GSdisplayhelpmessage(Savemsg = 20);
    GSdisablebutton("root", "number", DRAW);
    GSdisablebutton("root", "quest", DRAW);
    GSdisablebutton("root", "country", DRAW);
    if (Show_results)
    {
      GSgetfieldrect("Question", &x, &y, NULL, NULL, &color);
      GSsetcolor(color);
      GSsetfont(LARGE);
      GSsetcurrentnamedwindow("cityname");
      GSwwritenamedmsg(x+10, y+SMALL, "resultsclue");
    }
    NotInStepThru = FALSE;              /* Change state of program */
    step_thru();
    return;
  }
  if (Show_results && !Moving)
    show_summary_table();
  else
  {
    if (Show_results)    
      GSdisplayhelpmessage(Savemsg = 8);
    else
      GSdisplayhelpmessage(Savemsg = 13);
    show_final_map();
    GSenablebutton("root", "number", DRAW);
    GSenablebutton("root", "quest", DRAW);
    GSenablebutton("root", "country", DRAW);
  }
}


/***********************************************************************
Draw a city circle marker.
***********************************************************************/
void draw_city_circle(
  int x,
  int y,
  GSColor color,
  City_circle_pixels *replaced_pixels)
{
  if (replaced_pixels != NULL)  /* save pixel values for possible undo */
  {
    replaced_pixels->xy = GSwgetpoint(x, y);
    replaced_pixels->xyminus1 = GSwgetpoint(x, y-1);
    replaced_pixels->xplus1y = GSwgetpoint(x+1, y);
    replaced_pixels->xyplus1 = GSwgetpoint(x, y+1);
    replaced_pixels->xminus1y = GSwgetpoint(x-1, y);
    replaced_pixels->xyminus2 = GSwgetpoint(x, y-2);
    replaced_pixels->xplus1yminus1 = GSwgetpoint(x+1, y-1);
    replaced_pixels->xplus2y = GSwgetpoint(x+2, y);
    replaced_pixels->xplus1yplus1 = GSwgetpoint(x+1, y+1);
    replaced_pixels->xyplus2 = GSwgetpoint(x, y+2);
    replaced_pixels->xminus1yplus1 = GSwgetpoint(x-1, y+1);
    replaced_pixels->xminus2y = GSwgetpoint(x-2, y);
    replaced_pixels->xminus1yminus1 = GSwgetpoint(x-1, y-1);
  }
  GSsetcolor(color);
  GSwdrawpoint(x, y);
  GSwdrawpoint(x, y-1);
  GSwdrawpoint(x+1, y);
  GSwdrawpoint(x, y+1);
  GSwdrawpoint(x-1, y);
  GSwdrawpoint(x, y-2);
  GSwdrawpoint(x+1, y-1);
  GSwdrawpoint(x+2, y);
  GSwdrawpoint(x+1, y+1);
  GSwdrawpoint(x, y+2);
  GSwdrawpoint(x-1, y+1);
  GSwdrawpoint(x-2, y);
  GSwdrawpoint(x-1, y-1);
}

/***********************************************************************
Erase a city circle marker.
***********************************************************************/
void erase_city_circle(
  int x,
  int y,
  City_circle_pixels replaced_pixels)
{
  GSsetcolor(replaced_pixels.xy);
  GSwdrawpoint(x, y);
  GSsetcolor(replaced_pixels.xyminus1);
  GSwdrawpoint(x, y-1);
  GSsetcolor(replaced_pixels.xplus1y);
  GSwdrawpoint(x+1, y);
  GSsetcolor(replaced_pixels.xyplus1);
  GSwdrawpoint(x, y+1);
  GSsetcolor(replaced_pixels.xminus1y);
  GSwdrawpoint(x-1, y);
  GSsetcolor(replaced_pixels.xyminus2);
  GSwdrawpoint(x, y-2);
  GSsetcolor(replaced_pixels.xplus1yminus1);
  GSwdrawpoint(x+1, y-1);
  GSsetcolor(replaced_pixels.xplus2y);
  GSwdrawpoint(x+2, y);
  GSsetcolor(replaced_pixels.xplus1yplus1);
  GSwdrawpoint(x+1, y+1);
  GSsetcolor(replaced_pixels.xyplus2);
  GSwdrawpoint(x, y+2);
  GSsetcolor(replaced_pixels.xminus1yplus1);
  GSwdrawpoint(x-1, y+1);
  GSsetcolor(replaced_pixels.xminus2y);
  GSwdrawpoint(x-2, y);
  GSsetcolor(replaced_pixels.xminus1yminus1);
  GSwdrawpoint(x-1, y-1);
}
